# coding: utf-8

from typing import Union, Optional
import logging

from .meta_actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class ServoStar300(Actuator):
  """This class can drive a Kollmorgen ServoStar 300 servomotor conditioner in
  position, and set it to the analog or serial driving mode.

  It communicates with the servomotor over a serial connection. The
  :class:`~crappy.lamcube.Biaxe` Actuator can drive the same hardware, but only
  in speed.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Servostar to ServoStar300
  """

  def __init__(self,
               port: str,
               baudrate: int = 38400,
               mode: str = "serial") -> None:
    """Sets the instance attributes and initializes the parent class.

    Args:
      port: Path to the serial port used for communication.
      baudrate: The serial baud rate to use, as an :obj:`int`.
      mode: The driving mode to use when starting the test. Can be `'analog'`
        or `'serial'`. It can be changed afterward while the test is running,
        by sending the right command.
    
    .. deprecated:: 2.0.0 *device argument*
    .. versionchanged:: 2.0.0 use *port* instead of *device*
    """

    self._ser = None

    super().__init__()

    self._port = port
    self._mode = mode
    self._baudrate = baudrate

    self._last_pos = None

    if mode not in ('analog', 'serial'):
      raise AttributeError(f"No such mode: {mode}")

  def open(self) -> None:
    """Initializes the serial connection and sets the desired driving mode."""

    # Opening the serial connection
    self.log(logging.INFO, f"Opening the serial port {self._port} with "
                           f"baudrate {self._baudrate}")
    self._ser = serial.Serial(self._port, baudrate=self._baudrate, timeout=2)
    self.log(logging.DEBUG, f"Writing b'ANCNFG 0\\r\\n' to port {self._port}")
    self._ser.write('ANCNFG 0\r\n')

    # Setting the desired mode
    if self._mode == "analog":
      self._set_mode_analog()
    else:
      self._set_mode_serial()

    self.log(logging.DEBUG, f"Writing b'EN\\r\\n' to port {self._port}")
    self._ser.write('EN\r\n')
    self.log(logging.DEBUG, f"Writing b'MH\\r\\n' to port {self._port}")
    self._ser.write('MH\r\n')

  def set_position(self,
                   pos: Union[float, bool],
                   speed: Optional[float]) -> None:
    """Sets the target position for the motor.

    Also allows switching to the serial or analog driving mode if the target
    position is a :obj:`bool`. The acceleration and deceleration are set to
    `200`.

    Args:
      pos: The target position to reach, as a :obj:`float`. Alternatively, a
        value of :obj:`True` sets the driving mode to serial, and :obj:`False`
        sets the driving mode to analog.
      speed: The speed at which the actuator should reach its target position.
        If no speed is specified, the default is `20000`.
    
    .. deprecated:: 2.0.0 remove *acc* and *dec* arguments
    .. versionchanged:: 2.0.0 *speed* is now a mandatory argument
    """

    if speed is None:
      speed = 20000

    # Nothing to do if the command is the same as the previous
    if self._last_pos == pos:
      return

    # To allow setting the mode with a boolean command, so that the actuator
    # can be driven from a single generator
    if isinstance(pos, bool):
      if pos:
        self._set_mode_serial()
      else:
        self._set_mode_analog()
      return

    # The commands can only be set from Crappy in serial mode
    elif self._mode != "serial":
      self._set_mode_serial()

    # Writing the command to the actuator
    self._ser.flushInput()
    self.log(logging.DEBUG, f"Writing b'ORDER 0 {int(pos)} {speed} 8192 200 "
                            f"200 0 0 0 0\\r\\n' to port {self._port}")
    self._ser.write(f"ORDER 0 {int(pos)} {speed} 8192 200 200 0 0 0 0\r\n")
    self.log(logging.DEBUG, f"Writing b'MOVE 0\\r\\n' to port {self._port}")
    self._ser.write("MOVE 0\r\n")

    # Saving the last command
    self._last_pos = pos

  def get_position(self) -> Optional[float]:
    """Reads and returns the current position of the motor.

    .. versionchanged:: 1.5.2 renamed from get_pos to get_position
    """

    # Requesting a position reading
    self._ser.flushInput()
    self.log(logging.DEBUG, f"Writing b'PFB\\r\\n' to port {self._port}")
    self._ser.write("PFB\r\n")

    # Reading until getting the stop sequence, or the actuator stops responding
    r = ''
    while r != "PFB\r\n":
      # Keeping only the last 5 received characters
      if len(r) == 5:
        r = r[1:]
      # Reading the next character
      r += self._ser.read()
      # Aborting if no new character could be read
      if not r:
        self.log(logging.ERROR, "Timeout error ! Make sure the servostar is "
                                "on !")
        return

    # The next reading should give the position
    ret = self._ser.readline()
    self.log(logging.DEBUG, f"Read {ret} on port {self._port}")
    return int(ret)

  def stop(self) -> None:
    """Sends a command for stopping the motor."""

    if self._ser is not None:
      self.log(logging.DEBUG, f"Writing b'DIS\\r\\n' to port {self._port}")
      self._ser.write("DIS\r\n")
      self._ser.flushInput()

  def close(self) -> None:
    """Closes the serial connection."""

    if self._ser is not None:
      self.log(logging.INFO, f"Closing the serial port {self._port}")
      self._ser.close()

  def _set_mode_serial(self) -> None:
    """Sets the driving mode to serial."""

    self._ser.flushInput()
    self.log(logging.DEBUG, f"Writing b'OPMODE 8\\r\\n' to port {self._port}")
    self._ser.write('OPMODE 8\r\n')
    self._mode = "serial"

  def _set_mode_analog(self) -> None:
    """Sets the driving mode to analog."""

    self._last_pos = None
    self._ser.flushInput()
    self.log(logging.DEBUG, f"Writing b'OPMODE A\\r\\n' to port {self._port}")
    self._ser.write('OPMODE 1\r\n')
    self._mode = "analog"



# coding: utf-8

from typing import Optional
import logging
from  warnings import warn

from .meta_actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class SchneiderMDrive23(Actuator):
  """This class can drive Schneider Electric MDrive 23 stepper motor in speed
  and in position.

  It communicates with the motor over a serial connection.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from CM_drive to SchneiderMDrive23
  """

  def __init__(self,
               port: str = '/dev/ttyUSB0',
               baudrate: int = 9600) -> None:
    """Sets the instance attributes and initializes the parent class.

    Args:
      port: The path to the serial port to open for the serial connection.
      baudrate: The baudrate to use for serial communication.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._ser = None

    super().__init__()

    self._port = port
    self._baudrate = baudrate

  def open(self) -> None:
    """Opens the serial connection to the stepper motor."""

    self.log(logging.INFO, f"Opening the serial port {self._port} with "
                           f"baudrate {self._baudrate}")
    self._ser = serial.Serial(self._port, self._baudrate, timeout=0.1)

  def set_speed(self, speed: float) -> None:
    """Sets the target speed of the stepper motor.

    Args:
      speed: The target speed to set, in `mm/min`.
    """

    # Closing and reopening to get rid of errors
    self._ser.close()
    self._ser.open()

    # Sending the command only if it's below the maximum allowed value
    if abs(speed) < 1000000:
      self.log(logging.DEBUG, f"Writing b'SL {int(speed)}\\r' to port "
                              f"{self._port}")
      self._ser.write(f'SL {int(speed)}\r')
      self._ser.read(self._ser.inWaiting())
    else:
      self.log(logging.WARNING, "Maximum speed exceeded, not setting speed")

  def set_position(self, position: float, _: Optional[float]) -> None:
    """Sets the target position for the stepper motor.

    Args:
      position: The target position to reach, in `mm`.
      _: The speed argument is ignored.
    
    .. versionchanged:: 2.0.0 *speed* is now a mandatory argument
    .. deprecated:: 2.0.0 *motion_type* argument
    """

    # Closing and reopening to get rid of errors
    self._ser.close()
    self._ser.open()

    # Sending the position command
    self.log(logging.DEBUG, f"Writing b'MR {int(position)}\\r' to port "
                            f"{self._port}")
    self._ser.write(f'MR {int(position)}\r')
    self._ser.readline()

  def get_position(self) -> float:
    """Reads, displays and returns the current position in `mm`.
    
    .. versionchanged:: 1.5.2 renamed from get_pos to get_position
    """

    # Closing and reopening to get rid of errors
    self._ser.close()
    self._ser.open()

    # Asking for a position reading
    self.log(logging.DEBUG, f"Writing b'PR P \\r' to port {self._port}")
    self._ser.write('PR P \r')
    pfb = self._ser.readline()
    self.log(logging.DEBUG, f"Read {pfb} from port {self._port}")
    pfb1 = self._ser.readline()
    self.log(logging.DEBUG, f"Read {pfb1} from port {self._port}")

    return int(pfb1)

  def stop(self) -> None:
    """Sends a command for stopping the motor."""

    if self._ser is not None:
      # Closing and reopening to get rid of errors
      self._ser.close()
      self._ser.open()

      self.log(logging.DEBUG, f"Writing b'SL 0\\r' to port {self._port}")
      self._ser.write('SL 0\r')

  def close(self) -> None:
    """Close the serial connection."""

    if self._ser is not None:
      self.log(logging.INFO, f"Closing the serial port {self._port}")
      self._ser.close()



# coding: utf-8

import numpy as np
from typing import Dict, Any
import logging

from .meta_modifier import Modifier


class MovingAvg(Modifier):
  """Modifier replacing the data of each label with its average value over a
  chosen number of points.

  Unlike :class:`~crappy.modifier.Mean`, it returns a value each time data is
  received from the upstream Block.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Moving_avg to MovingAvg
  """

  def __init__(self, n_points: int = 100) -> None:
    """Sets the args and initializes the parent class.

    Args:
      n_points: The maximum number of points on which to compute the average.
    
    .. versionchanged:: 1.5.10 renamed *npoints* argument to *n_points*
    """

    super().__init__()
    self._n_points = n_points
    self._buf = None

  def __call__(self, data: Dict[str, Any]) -> Dict[str, Any]:
    """Receives data from the upstream Block, computes the average of every
    label and replaces the original data with it.
    
    .. versionchanged:: 2.0.0 renamed from evaluate to __call__
    """

    self.log(logging.DEBUG, f"Received {data}")

    # Initializing the buffer
    if self._buf is None:
      self._buf = {key: [value] for key, value in data.items()}

    ret = {}
    for label in data:
      # Updating the buffer with the newest data
      self._buf[label].append(data[label])

      # Trimming the buffer if there's too much data
      if len(self._buf[label]) > self._n_points:
        self._buf[label] = self._buf[label][-self._n_points:]

      # Calculating the average for each label
      try:
        ret[label] = np.mean(self._buf[label])
      except TypeError:
        ret[label] = self._buf[label][-1]

    self.log(logging.DEBUG, f"Sending {ret}")
    return ret



# coding: utf-8

from time import time
from numpy import loadtxt, interp
from typing import Dict, Union
import pathlib
import logging

from .meta_path import Path


class Custom(Path):
  """Generates a custom Path from a text file, until the file is exhausted.

  The file can be in any text format, including the most common `.csv` and
  `.txt` extensions.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               file_name: Union[str, pathlib.Path],
               delimiter: str = ',') -> None:
    """Loads the file and sets the arguments.

    The stop condition is simply to reach the last timestamp given in the
    file.

    Args:
      file_name: Path to the file containing the information on the Generator
        Path. Can be either a :obj:`str` or a :obj:`pathlib.Path`. The file
        must contain two columns: the first one containing timestamps (starting
        from 0), the other one containing the values.
      delimiter: The delimiter between columns in the file, usually a coma.
    
    .. versionchanged:: 1.5.10 renamed *time* argument to *_last_time*
    .. versionchanged:: 1.5.10 renamed *cmd* argument to *_last_cmd*
    .. versionchanged:: 2.0.0 renamed *filename* argument to *file_name*
    .. versionremoved:: 2.0.0 *_last_time* and *_last_cmd* arguments
    """

    super().__init__()

    self.log(logging.DEBUG, f"Extracting data from file {file_name}")
    array = loadtxt(pathlib.Path(file_name), delimiter=delimiter)

    if array.shape[1] != 2:
      raise ValueError(f'The file {file_name} should contain exactly two'
                       f'columns !')

    self._timestamps = array[:, 0]
    self._values = array[:, 1]

  def get_cmd(self, _: Dict[str, list]) -> float:
    """Returns the value to send or raises :exc:`StopIteration` if the stop
    condition is met.

    The value is interpolated from the given file.
    """

    t = time()
    if t - self.t0 > self._timestamps[-1]:
      self.log(logging.DEBUG, "Stop condition met")
      raise StopIteration
    return float(interp(t - self.t0, self._timestamps, self._values))



# coding: utf-8

from time import time
from typing import Union, Dict, Optional
import logging

from .meta_path import Path, ConditionType


class Ramp(Path):
  """Sends a ramp signal varying linearly over time, until the stop condition
  is met.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               condition: Union[str, ConditionType],
               speed: float,
               init_value: Optional[float] = None):
    """Sets the arguments and initializes the parent class.

    Args:
      condition: The condition for switching to the next Path. Refer to
        :class:`~crappy.blocks.generator_path.meta_path.Path` for more
        information.
      speed: The slope of the ramp, in `units/s`.
      init_value: If given, overwrites the last value of the signal as the
        starting point for the ramp. In the specific case when this path is the
        first one in the Generator Paths, this argument must be given !
    
    .. versionchanged:: 1.5.10 renamed *time* argument to *_last_time*
    .. versionchanged:: 1.5.10 renamed *cmd* argument to *_last_cmd*
    .. versionadded:: 1.5.10 *init_value* argument
    .. versionremoved:: 2.0.0 *_last_time* and *_last_cmd* arguments
    """

    super().__init__()

    if init_value is None and self.last_cmd is None:
      raise ValueError('For the first path, an init_value must be given !')

    # Setting the attributes
    self._condition = self.parse_condition(condition)
    self._speed = speed
    self._init_value = init_value if init_value is not None else self.last_cmd

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Returns the value to send or raises :exc:`StopIteration` if the stop
    condition is met."""

    # Checking if the stop condition is met
    if self._condition(data):
      self.log(logging.DEBUG, "Stop condition met")
      raise StopIteration

    # Returning the current value
    return self._init_value + (time() - self.t0) * self._speed



# coding: utf-8

from time import time, sleep
from typing import Tuple, List, Optional
from numpy import ndarray
from subprocess import run
from re import findall, search
import logging

from .meta_camera import Camera
from ._v4l2_base import V4L2Helper
from .._global import OptionalModule

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class CameraOpencv(Camera, V4L2Helper):
  """A class for reading images from any camera able to interface with OpenCV.

  The number of the video device to read images from can be specified.

  This camera class is less performant than the
  :class:`~crappy.camera.CameraGstreamer` one that relies on GStreamer, but the
  installation of OpenCV is way easier than the one of GStreamer, especially on
  Windows !

  Warning:
    There are two classes for CameraOpencv, one for Linux based on
    `v4l-utils`, and another one for Linux (without `v4l-utils`) and other OS.
    Depending on the installation of `v4l-utils` and the OS, the correct class
    will be automatically imported. The version using `v4l-utils` allows tuning
    more parameters than the basic version.

  .. versionadded:: 1.5.9
  .. versionchanged:: 2.0.0 renamed from Camera_opencv to CameraOpencv
  """

  def __init__(self) -> None:
    """Sets variables and adds the channels setting."""

    Camera.__init__(self)
    V4L2Helper.__init__(self)

    self._cap = None
    self._device_num: Optional[int] = None
    self._formats: List[str] = list()

  def open(self, device_num: int = 0, **kwargs) -> None:
    """Opens the video stream and sets any user-specified settings.

    Args:
      device_num: The index of the device to open, as an :obj:`int`.
      **kwargs: Any additional setting to set before opening the configuration
        window.
    """

    if not isinstance(device_num, int) or device_num < 0:
      raise ValueError("device_num should be an integer !")

    # Opening the videocapture device
    self.log(logging.INFO, "Opening the image stream from the camera")
    self._cap = cv2.VideoCapture(device_num)
    self._device_num = device_num

    # Getting the available formats for the selected device and filtering the
    # supported ones
    self._get_available_formats(device_num)
    supported = ('MJPG', 'YUYV')
    unavailable = set([_format.split()[0] for _format in self._formats
                       if _format.split()[0] not in supported])
    if unavailable:
      self.log(logging.WARNING, f"The formats {', '.join(unavailable)} "
                                f"are available but not implemented in Crappy")
    self._formats = [_format for _format in self._formats
                     if _format.split()[0] in supported]

    # Instantiating the format setting if there are formats left
    if self._formats:
      self.add_choice_setting(name='format',
                              choices=tuple(self._formats),
                              getter=self._get_format_size,
                              setter=self._set_format)

    # Getting the available parameters for the camera
    self._get_param(device_num)

    # Creating the different settings
    for param in self._parameters:
      if param.type == 'int':
        self.add_scale_setting(
          name=param.name,
          lowest=int(param.min),
          highest=int(param.max),
          getter=self._add_scale_getter(param.name, self._device_num),
          setter=self._add_setter(param.name, self._device_num),
          default=param.default,
          step=int(param.step))

      elif param.type == 'bool':
        self.add_bool_setting(
          name=param.name,
          getter=self._add_bool_getter(param.name, self._device_num),
          setter=self._add_setter(param.name, self._device_num),
          default=bool(int(param.default)))

      elif param.type == 'menu':
        if param.options:
          self.add_choice_setting(
            name=param.name,
            choices=param.options,
            getter=self._add_menu_getter(param.name, self._device_num),
            setter=self._add_setter(param.name, self._device_num),
            default=param.default)

      else:
        self.log(logging.ERROR, f'The type {param.type} is not yet'
                                f' implemented. Only int, bool and menu '
                                f'type are implemented. ')
        raise NotImplementedError

    self.add_choice_setting(name="channels", choices=('1', '3'), default='1')

    # Adding the software ROI selection settings
    if 'format' in self.settings:
      width, height = search(r'(\d+)x(\d+)', self._get_format_size()).groups()
      self.add_software_roi(int(width), int(height))

    # Setting the kwargs if any
    self.set_all(**kwargs)

  def get_image(self) -> Tuple[float, ndarray]:
    """Grabs a frame from the videocapture object and returns it along with a
    timestamp."""

    # Grabbing the frame and the timestamp
    t = time()
    ret, frame = self._cap.read()

    # Checking the integrity of the frame
    if not ret:
      raise IOError("Error reading the camera")

    # Returning the image in the right format, and its timestamp
    if self.channels == '1':
      return t, self.apply_soft_roi(cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY))
    else:
      return t, self.apply_soft_roi(frame)

  def close(self) -> None:
    """Releases the videocapture object."""

    if self._cap is not None:
      self.log(logging.INFO, "Closing the image stream from the camera")
      self._cap.release()

  def _get_width(self) -> int:
    """Returns the current image width."""

    return self._cap.get(cv2.CAP_PROP_FRAME_WIDTH)

  def _get_height(self) -> int:
    """Returns the current image height."""

    return self._cap.get(cv2.CAP_PROP_FRAME_HEIGHT)

  def _set_width(self, width: int) -> None:
    """Tries to set the image width."""

    self._cap.set(cv2.CAP_PROP_FRAME_WIDTH, width)

    # Reloading the software ROI selection settings
    if self._soft_roi_set:
      sleep(0.1)
      width, height = self._get_width(), self._get_height()
      self.reload_software_roi(width, height)

  def _set_height(self, height: int) -> None:
    """Tries to set the image height."""

    self._cap.set(cv2.CAP_PROP_FRAME_HEIGHT, height)

    # Reloading the software ROI selection settings
    if self._soft_roi_set:
      sleep(0.1)
      width, height = self._get_width(), self._get_height()
      self.reload_software_roi(width, height)

  def _set_format(self, img_format: str) -> None:
    """Sets the format of the image according to the user's choice."""

    # The format is made of a name, a size and a framerate
    format_name, img_size, fps = findall(r"(\w+)\s(\w+)\s\((\d+.\d+) fps\)",
                                         img_format)[0]

    # Setting the format
    self._cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*format_name))

    # Getting the width and height from the second half of the string
    width, height = map(int, img_size.split('x'))

    # Setting the size
    self._set_width(width)
    self._set_height(height)

    # Setting the acquisition frequency
    self._cap.set(cv2.CAP_PROP_FPS, float(fps))

    # Reloading the software ROI selection settings
    if self._soft_roi_set:
      sleep(0.1)
      width, height = search(r'(\d+)x(\d+)', self._get_format_size()).groups()
      self.reload_software_roi(int(width), int(height))

  def _get_format_size(self) -> str:
    """Parses the ``v4l2-ctl -all`` command to get the current image format as
    a :obj:`str`."""

    # Sending the v4l2-ctl command
    command = ['v4l2-ctl', '-d', str(self._device_num), '--all']
    self.log(logging.DEBUG, f"Getting the current image formats with "
                            f"command {' '.join(command)}")
    ret = run(command, capture_output=True, text=True).stdout
    self.log(logging.DEBUG, f"Got the following image formats: {ret}")

    # Parsing the answer
    format_ = width = height = fps = ''
    if search(r"Pixel Format\s*:\s*'(\w+)'", ret) is not None:
      format_, *_ = search(r"Pixel Format\s*:\s*'(\w+)'", ret).groups()
    if search(r"Width/Height\s*:\s*(\d+)/(\d+)", ret) is not None:
      width, height = search(r"Width/Height\s*:\s*(\d+)/(\d+)", ret).groups()
    if search(r"Frames per second\s*:\s*(\d+.\d+)", ret) is not None:
      fps, *_ = search(r"Frames per second\s*:\s*(\d+.\d+)", ret).groups()

    return f'{format_} {width}x{height} ({fps} fps)'



# coding: utf-8

from typing import Union

from ..meta_inout import InOut
from ...tool.ft232h import FT232HServer as FT232H, ft232h_pin_nr, USBArgsType


class GPIOSwitchFT232H(InOut):
  """This class can drive a GPIO high or low on a FT232H.

  It is similar to the :class:`~crappy.inout.GPIOSwitch` class, except this
  class is specific for use with an :class:`~crappy.tool.ft232h.FT232H` USB to
  GPIO converter.

  When the command value is `1` the GPIO is turned high, when the command is
  `0` it is turned low. Any value other than `0` and `1` raises an error.
  
  .. versionadded:: 2.0.0
  """

  ft232h = True

  def __init__(self,
               pin_out: Union[int, str],
               _ft232h_args: USBArgsType = tuple()) -> None:
    """Checks the validity of the arguments.

    Args:
      pin_out: The GPIO pin to be controlled. On Raspberry Pi, should be an
        integer corresponding to a GPIO in BCM convention. On FT232H, should be
        a string corresponding to the name of a GPIO. With the `'blinka'`
        backend, should be a string holding the name of the pin. Refer to
        blinka's specific documentation for each board for more information.
      _ft232h_args: This argument is meant for internal use only and should not
        be provided by the user. It contains the information necessary for
        setting up the FT232H.
    """

    self._ft232h = None
    self._pin_out = None

    # Checking that the given pin is valid
    if pin_out not in ft232h_pin_nr:
      raise TypeError(f'{pin_out} is not a valid pin for the ft232h backend !')

    super().__init__()

    (block_index, block_lock, command_file, answer_file, shared_lock,
     current_block) = _ft232h_args

    self._ft232h = FT232H(mode='GPIO_only',
                          block_index=block_index,
                          current_block=current_block,
                          command_file=command_file,
                          answer_file=answer_file,
                          block_lock=block_lock,
                          shared_lock=shared_lock)

    self._pin_out = pin_out

  def set_cmd(self, *cmd: int) -> None:
    """Drives the GPIO according to the command.

    Args:
      cmd: 1 for driving the GPIO high, 0 for driving it low.
    """

    if cmd[0] not in [0, 1]:
      raise ValueError("The GPIO input can only be 0 or 1")

    self._ft232h.set_gpio(self._pin_out, cmd[0])

  def close(self) -> None:
    """Releases the GPIO."""

    if self._ft232h is not None:
      self._ft232h.close()



# coding: utf-8

from typing import Optional, Callable, Tuple
from itertools import zip_longest
import logging

from .camera_setting import CameraSetting


class CameraChoiceSetting(CameraSetting):
  """Camera setting that can take any value from a predefined list of
  values.

  It is a child of
  :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting`.
  
  .. versionadded:: 1.5.10
  .. versionchanged:: 2.0.0
     renamed from Camera_choice_setting to CameraChoiceSetting
  """

  def __init__(self,
               name: str,
               choices: Tuple[str, ...],
               getter: Optional[Callable[[], str]] = None,
               setter: Optional[Callable[[str], None]] = None,
               default: Optional[str] = None) -> None:
    """Sets the attributes.

    Args:
      name: The name of the setting, that will be displayed in the GUI.
      choices: A tuple listing the possible values for the setting.
      getter: The method for getting the current value of the setting.
      setter: The method for setting the current value of the setting.
      default: The default value to assign to the setting.
    """

    self.choices = choices

    if default is None:
      default = choices[0]

    super().__init__(name, getter, setter, default)

    self.tk_obj = list()

  def reload(self,
             choices: Tuple[str, ...],
             default: Optional[str] = None) -> None:
    """Allows modifying the choices of the radio buttons once they have been
    instantiated.

    As the layout of the GUI is already fixed, the number of displayed options
    cannot vary. It is thus not possible to propose more choices than those
    initially proposed. Reversely, if fewer new options ar proposed then some
    radio buttons won't be affected a value.
    
    .. versionadded:: 2.0.0
    """

    self.log(logging.DEBUG, f"Reloading the setting {self.name}")

    # Updating the default value
    if default is not None:
      self.default = default
    else:
      self.default = choices[0]

    # Updating the radio buttons and the setting value
    if self.tk_obj:
      for button, choice in zip_longest(self.tk_obj, choices):
        # If there are more choices than buttons, ignoring the extra choices
        if button is None:
          self.log(logging.WARNING,
                   f"Too many choices given when reloading the {self.name} "
                   f"setting, ignoring the extra ones")
          break
        # If there are more buttons than choices, disabling the extra buttons
        if choice is None:
          self.log(logging.WARNING,
                   f"Too few choices given when reloading the {self.name} "
                   f"setting, disabling the extra buttons")
          button.configure(state='disabled', value='', text='')

        # Updating the text and value of the button, and enabling it
        button.configure(value=choice, text=choice, state='normal')
    if self.tk_var is not None:
      self.tk_var.set(self.value)



# coding: utf-8

from time import sleep
from typing import Optional
import logging
from multiprocessing import current_process

from .meta_actuator import MetaActuator


class Actuator(metaclass=MetaActuator):
  """The base class for all Actuator classes, allowing to keep track of them
  and defining methods shared by all of them.

  The Actuator objects are helper classes used by the
  :class:`~crappy.blocks.Machine` Block to communicate with motors or other
  actuators.

  .. versionadded:: 1.4.0
  """

  ft232h: bool = False

  def __init__(self, *_, **__) -> None:
    """Initializes the instance attributes."""

    self._logger: Optional[logging.Logger] = None

  def log(self, level: int, msg: str) -> None:
    """Records log messages for the Actuator.

    Also instantiates the logger when logging the first message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.

    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{type(self).__name__}")

    self._logger.log(level, msg)

  def open(self) -> None:
    """This method should perform any action that's required for initializing
    the hardware and the communication with it.

    Communication with hardware should be avoided in the :meth:`__init__`
    method, and this method is where it should start. This method is called
    after Crappy's processes start, i.e. when the associated
    :class:`~crappy.blocks.Machine` already runs separately from all the other
    Blocks.

    It is fine for this method not to perform anything.

    .. versionadded:: 2.0.0
    """

    ...

  def set_speed(self, speed: float) -> None:
    """This method should drive the actuator so that it reaches the desired
    speed.

    It is used if the ``mode`` given in the ``actuators`` argument of the
    :class:`~crappy.blocks.Machine` Block is ``'speed'``.

    The value passed as the ``speed`` argument will be that received over
    the ``cmd_label`` given in the ``actuators`` argument of the
    :class:`~crappy.blocks.Machine` Block.

    Args:
      speed: The speed to reach, as a :obj:`float`.

    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING, f"The set_speed method was called but is not "
                              f"defined ! No command sent to the actuator.")
    sleep(1)

  def set_position(self, position: float, speed: Optional[float]) -> None:
    """This method should drive the actuator so that it reaches the desired
    position.

    It is used if the ``mode`` given in the ``actuators`` argument of the
    :class:`~crappy.blocks.Machine` Block is ``'position'``.

    The value passed as the ``position`` argument will be that received over
    the ``cmd_label`` given in the ``actuators`` argument of the
    :class:`~crappy.blocks.Machine` Block.

    The value passed as the ``speed`` argument will be either :

    * :obj:`None` if no ``speed`` nor ``speed_cmd_label`` were specified in the
      ``actuators`` argument of the :class:`~crappy.blocks.Machine` Block.

    * The value given with the ``speed`` key of the ``actuators`` argument
      of the :class:`~crappy.blocks.Machine` Block, if no other speed command
      was received in the meantime.

    * The last value received over the ``speed_cmd_label`` if it was set in the
      ``actuators`` argument of the :class:`~crappy.blocks.Machine` Block.
      Before the first speed command is received, the value will either be
      :obj:`None` if no ``speed`` was specified, else the value given as
      ``speed`` in the ``actuators`` argument of the
      :class:`~crappy.blocks.Machine` Block.

    Important:
      The ``speed`` value might be :obj:`None`, but it is not optional ! When
      writing a custom Actuator, make sure to always handle it.

    Args:
      position: The position to reach, as a :obj:`float`.
      speed: The speed at which to move to the desired position, as a
        :obj:`float`, or :obj:`None` if no speed is specified.

    .. versionadded:: 1.5.10
    .. versionchanged:: 2.0.0
       *speed* is now a mandatory argument even if it is :obj:`None`
    """

    self.log(logging.WARNING, f"The set_position method was called but is not "
                              f"defined ! No command sent to the actuator.")
    sleep(1)

  def get_speed(self) -> Optional[float]:
    """This method should return the current speed of the actuator, as a
    :obj:`float`.

    This speed will be sent to downstream Blocks, over the label given with
    the ``speed_label`` key of the ``actuators`` argument of the
    :class:`~crappy.blocks.Machine` Block.

    It is also fine for this method to return :obj:`None` if the speed could
    not be acquired.

    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING, f"The get_speed method as called but is not "
                              f"defined ! Define such a method, don't set the "
                              f"speed_label, or remove the output links of "
                              f"this block.")
    sleep(1)
    return

  def get_position(self) -> Optional[float]:
    """This method should return the current position of the actuator, as a
    :obj:`float`.

    This position will be sent to downstream Blocks, over the label given with
    the ``position_label`` key of the ``actuators`` argument of the
    :class:`~crappy.blocks.Machine` Block.

    It is also fine for this method to return :obj:`None` if the position could
    not be acquired.

    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING, f"The get_position method as called but is not "
                              f"defined ! Define such a method, don't set the "
                              f"position_label, or remove the output links of "
                              f"this block.")
    sleep(1)
    return

  def stop(self) -> None:
    """This method should stop all movements of the actuator, and if possible
    make sure that the actuator cannot move anymore.

    That includes for example de-energizing the actuator, or switching it to a
    locked state.

    This method will only be called once at the very end of the test. The
    default behavior is to call :meth:`set_speed` to set the speed to `0`. This
    method doesn't need to be overriden if the actuator doesn't have any
    feature for stopping other than speed control.

    .. versionadded:: 2.0.0
    """

    self.set_speed(0)

  def close(self) -> None:
    """This method should perform any action required for properly ending the
    test and closing the communication with hardware.

    It will be called when the associated :class:`~crappy.blocks.Machine`
    receives the order to stop (usually because the user hit `CTRL+C`, or
    because a :class:`~crappy.blocks.Generator` Block reached the end of its
    path, or because an exception was raised in any of the Blocks).

    It is fine for this method not to perform anything.

    .. versionadded:: 2.0.0
    """

    ...



# coding: utf-8

from numpy import ndarray
import logging
from multiprocessing import current_process
from typing import Optional


class Overlay:
  """This class is the base class for all the classes adding overlays on top of
  the images displayed by a :class:`~crappy.blocks.camera_processes.Displayer`
  Process of a :class:`~crappy.blocks.Camera` Block.

  Also used for drawing overlays on top of the images in the
  :class:`~crappy.tool.camera_config.CameraConfig` window, for the children of
  the Camera Block supporting it.

  It is mainly useful for providing the :meth:`log` method, and creating a
  clear architecture. It is also relevant to use for type-hinting.

  .. versionadded:: 2.0.0
  """

  def __init__(self) -> None:
    """Simply initializes the logger to :obj:`None`."""

    self._logger: Optional[logging.Logger] = None

  def draw(self, img: ndarray) -> None:
    """This method takes the image to display as an input, draws an overlay on
    top of it, and returns the modified image.

    It is meant to be overriden by subclasses of this class.
    """

    ...

  def log(self, log_level: int, msg: str) -> None:
    """Method for recording log messages from the Overlay class.

    Args:
      log_level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    """

    if self._logger is None:
      self._logger = logging.getLogger(f"{current_process().name}."
                                       f"{type(self).__name__}")

    self._logger.log(log_level, msg)



# coding: utf-8

from struct import unpack
from time import time
from typing import Optional, Dict, Callable, Iterable, Union
import logging

from .meta_block import Block
from .._global import OptionalModule

try:
  from serial import Serial
  from serial.serialutil import SerialException
except (ModuleNotFoundError, ImportError):
  Serial = OptionalModule("pyserial")
  SerialException = OptionalModule("pyserial")


class UController(Block):
  """Block for interfacing over serial with an external device, written mostly
  for communication with microcontrollers.

  It can send labeled commands to the device, and/or receive labeled data from
  it. This Block is meant to be used along with the `microcontroller.py`
  MicroPython template located in the `tool` folder of Crappy, even though it
  is not mandatory. A given syntax needs to be followed for any data to be
  exchanged.
  
  .. versionadded:: 1.5.8
  """

  def __init__(self,
               labels: Optional[Union[str, Iterable[str]]] = None,
               cmd_labels: Optional[Union[str, Iterable[str]]] = None,
               init_output: Optional[Dict[str, float]] = None,
               post_process: Optional[Dict[str,
                                           Callable[[float], float]]] = None,
               t_device: bool = False,
               port: str = '/dev/ttyUSB0',
               baudrate: int = 115200,
               display_freq: bool = False,
               freq: Optional[float] = 100,
               debug: Optional[bool] = False) -> None:
    """Checks the validity of the arguments.

    Args:
      labels: An iterable (like a :obj:`list` or a :obj:`tuple`) containing the
        labels to get from the device (as :obj:`str`). Only these labels should
        be given as argument to the :meth:`send_to_pc` method in the
        MicroPython script. If this argument is not :obj:`None`, then the
        ``init_output`` argument should be given as well. No more than 9 labels
        should be given. If there's only one label to acquire, it can be given
        directly as a :obj:`str` and not in an iterable.
      cmd_labels: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        the command labels that will be sent to the device upon reception from
        an upstream Block. The variables in the MicroPython script should have
        these exact names. Not more than 9 cmd_labels should be given. If
        there's only one command label, it can be given directly as a
        :obj:`str` and not in an iterable.
      init_output: If the ``labels`` argument is not :obj:`None`, the values to 
        output to downstream Blocks for each label as long as no value has been 
        received from the device. An initial output value must be given for 
        each label.
      post_process: Optionally allows applying a function to the data of a 
        label before transmitting it to downstream Blocks. It is possible to 
        give functions for only part of the labels.
      t_device: It :obj:`True`, the timestamp returned under the label `'t(s)'` 
        is the one of the device, not the one of Crappy. It may reduce the 
        maximum achievable sample rate, as more bytes have to be transmitted, 
        but it is also far more precise.
      port: The serial port to open for communicating with the device. In 
        Windows, they are usually called `COMx`, whereas in Linux and Mac 
        they're called `/dev/ttyxxxx`.
      baudrate: The baudrate for serial communication. It depends on the 
        capabilities of the device.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionadded:: 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    self._bus = None

    super().__init__()
    self.debug = debug

    if not isinstance(display_freq, bool):
      raise TypeError("display_freq should be either True or False !")
    self.display_freq = display_freq

    if not isinstance(freq, float) and not isinstance(freq, int) or freq <= 0:
      raise TypeError("freq should be a positive float !")
    self.freq = freq

    if not isinstance(t_device, bool):
      raise TypeError("t_device should be either True or False !")
    self._t_device = t_device

    if not isinstance(port, str):
      raise TypeError("port should be a string !")
    self._port = port

    if not isinstance(baudrate, int) or baudrate < 0:
      raise ValueError("baudrate should be a positive integer !")
    self._baudrate = baudrate

    # Forcing the labels into a list
    if labels is not None and isinstance(labels, str):
      self._labels = [labels]
    elif labels is not None:
      self._labels = list(labels)
    else:
      self._labels = None

    # Forcing the cmd_labels into a list
    if cmd_labels is not None and isinstance(cmd_labels, str):
      self._cmd_labels = [cmd_labels]
    elif cmd_labels is not None:
      self._cmd_labels = list(cmd_labels)
    else:
      self._cmd_labels = None

    if self._labels is not None and len(self._labels) > 9:
      raise ValueError("Sorry, a maximum of 9 labels is allowed !")
    if self._cmd_labels is not None and len(self._cmd_labels) > 9:
      raise ValueError("Sorry, a maximum of 9 cmd_labels is allowed !")

    if self._cmd_labels is not None:
      self._prev_cmd = {cmd_label: None for cmd_label in self._cmd_labels}

    if init_output is not None and not isinstance(init_output, dict):
      raise TypeError("init_output should be a dict !")
    if self._labels is not None and not all(
        label in (init_output if init_output is not None else dict())
        for label in self._labels):
      raise ValueError("Every label should have an init_output value !")
    self._out = init_output

    if post_process is not None and (not isinstance(post_process, dict) or
                                     not all(callable(func) for func
                                             in post_process.values())):
      raise TypeError("post_process should be a dict of callables !")
    self._post_process = post_process if post_process is not None else {}

    self._buffer = None
    self._cmd_table = None
    self._labels_table = None

  def prepare(self) -> None:
    """Opens the serial port, and sends a `'go'` message to the device.

    Also shares with the device two tables associating each `cmd_label` and
    `label` with an integer. This allows reducing the traffic on the serial
    bus.

    Note:
      The commands are sent as text because some boards cannot read bytes from
      the `stdin` buffer in MicroPython.
    """

    # Checking if the link layout is relevant with respect to the arguments
    if self._labels is not None and not self.outputs:
      raise IOError("labels are specified but there's no output link !")
    if self._cmd_labels is not None and not self.inputs:
      raise IOError("cmd_labels are specified but there's no input link !")

    # Buffer for storing the received bytes
    if self._labels is not None:
      self._buffer = b''

    # Opening the serial port
    try:
      self.log(logging.INFO, f"Opening the serial port {self._port} with "
                             f"baudrate {self._baudrate}")
      self._bus = Serial(self._port,
                         self._baudrate,
                         timeout=0,
                         write_timeout=0)
    except SerialException:
      raise IOError(f"Couldn't connect to the device on the port {self._port}")

    # Assigning indexes to the cmd_labels and labels, to identify them easily
    # and reduce the traffic on the bus
    if self._cmd_labels is not None:
      self._cmd_table = {label: i for i, label
                         in enumerate(self._cmd_labels, start=1)}
    else:
      self._cmd_table = dict()
    self.log(logging.DEBUG, f"Command table : {self._cmd_table}")

    if self._labels is not None:
      self._labels_table = {label: i for i, label
                            in enumerate(self._labels, start=1)}
    else:
      self._labels_table = dict()

    # The presence of the label 't(s)' indicates that the device should return
    # a timestamp along with the data
    if self._labels is not None and self._t_device:
      self._labels_table.update({'t(s)': 0})

    self.log(logging.DEBUG, f"Labels table : {self._labels_table}")

    # Emptying the read buffer before starting
    try:
      self._bus.reset_input_buffer()
      self._bus.reset_output_buffer()
    except SerialException:
      raise IOError(f"Reading from the device on port {self._port} failed, "
                    f"it may have been disconnected.")

    # Sending the 'go' command to start the device
    self.log(logging.INFO, f"Sending start command on port {self._port}")
    try:
      msg = b''.join((b'go', str(len(self._cmd_table)).encode(),
                      str(len(self._labels_table)).encode(), b'\r\n'))
      self._bus.write(msg)
      self.log(logging.DEBUG, f"Sent {msg} on the port {self._port}")
    except SerialException:
      raise IOError(f"Writing to the device on port {self._port} failed, "
                    f"it may have been disconnected.")

    # Sending the table of cmd_labels and their indexes
    self.log(logging.INFO, f"Sending the command labels table on port "
                           f"{self._port}")
    for cmd, i in self._cmd_table.items():
      try:
        msg = b''.join((str(i).encode(), cmd.encode(), b'\r\n'))
        self._bus.write(msg)
        self.log(logging.DEBUG, f"Sent {msg} on the port {self._port}")
      except SerialException:
        raise IOError(f"Writing to the device on port {self._port} failed, "
                      f"it may have been disconnected.")

    # Sending the table of labels and their indexes
    self.log(logging.INFO, f"Sending the labels table on port {self._port}")
    for label, i in self._labels_table.items():
      try:
        msg = b''.join((str(i).encode(), label.encode(), b'\r\n'))
        self._bus.write(msg)
        self.log(logging.DEBUG, f"Sent {msg} on the port {self._port}")
      except SerialException:
        raise IOError(f"Writing to the device on port {self._port} failed, "
                      f"it may have been disconnected.")

  def loop(self) -> None:
    """First sends the commands from upstream Blocks to the device, then reads
    the data from the device and sends it to the downstream Blocks.

    Important:
      The precision of the commands sent to the device is limited to 3 digits
      after the decimal point, to limit the traffic on the bus. Adapt the range
      of the command values consequently.

    Note:
      Commands are sent as text, because some boards cannot read bytes from the
      `stdin` buffer in MicroPython. Data is however received on the PC from
      the device as bytes.
    """

    """Loop for sending the commands to the device.
    
    First, for each label in the cmd_labels list we search for a matching 
    command in the upstream links. Only the first one found is considered. If 
    the command value is different from the last value of this label, then the 
    command is sent to the device. Otherwise it is ignored. The command value
    is then stored as the last value of the label.
    """
    if self._cmd_labels is not None:
      cmd = self.recv_last_data()
      for label in self._cmd_labels:
        if label in cmd and cmd[label] != self._prev_cmd[label]:
          msg = b''.join((f'{self._cmd_table[label]}'
                          f'{cmd[label]:.3f}'.encode(),
                          b'\r\n'))

          # Information for debugging
          self.log(logging.DEBUG, f"Sending {msg} on the port {self._port}")

          # Sending the actual message to the device
          try:
            self._bus.write(msg)
          except SerialException:
            raise IOError(f"Writing to the device on port {self._port} "
                          f"failed, it may have been disconnected.")
          self._prev_cmd[label] = cmd[label]

    """Loop for receiving data from the device.
    
    A given number of bytes is read from the device, depending if `t_device` is
    True or False. they are then parsed to extract the index of the label, the
    data, and optionally the timestamp. If a label with a matching index is
    present in labels, its value is updated, as well as the timestamp. The
    values of ALL the labels are then sent to downstream blocks.
    """
    if self._labels is not None:
      # Reading the message from the device
      retries = 3
      while retries:
        try:
          recv = self._bus.read()
        except SerialException:
          raise IOError(f"Reading from the device on port {self._port} "
                        f"failed, it may have been disconnected.")

        # This prevents the loop from staying stuck with an incomplete reading
        if not recv:
          retries -= 1
          continue

        self._buffer += recv
        # Exiting the loop when the desired number of bytes is reached
        if (self._t_device and len(self._buffer) == 9) or \
            (not self._t_device and len(self._buffer) == 5):
          break

      # There was no message from the device, or it was incomplete
      if (self._t_device and len(self._buffer) != 9) or \
          (not self._t_device and len(self._buffer) != 5):
        return

      # Information for debugging
      self.log(logging.DEBUG, f"Received {self._buffer} on the port "
                              f"{self._port}")

      # Parsing the received bytes
      read = unpack('<ibf' if self._t_device else '<bf', self._buffer)
      self._buffer = b''

      # Reading the timestamp if relevant
      if self._t_device:
        self._out['t(s)'] = read[0] / 1000
        read = read[1:]

      self.log(logging.DEBUG, f"Read value {read} from the device")

      # Updating the label value and sending to the downstream blocks
      for label in self._labels:
        if read[0] == self._labels_table[label]:
          value = read[1]
          self._out[label] = self._post_process[label](value) if \
              label in self._post_process else value
          if not self._t_device:
            self._out['t(s)'] = time() - self.t0
          self.send(self._out)
          break

    # Emptying the port buffer even if the messages are not processed
    else:
      try:
        self._bus.reset_input_buffer()
      except SerialException:
        raise IOError(f"Reading from the device on port {self._port} "
                      f"failed, it may have been disconnected.")

  def finish(self) -> None:
    """Closes the serial port, and sends a `'stop!'` message to the device."""

    if self._bus is not None:
      # Sending a 'stop!' message to the device
      self.log(logging.INFO, f"Sending stop command on port {self._port}")
      try:
        msg = b'stop!\r\n'
        self._bus.write(msg)
        self.log(logging.DEBUG, f"Sent {msg} on the port {self._port}")
      except SerialException:
        pass

      self.log(logging.INFO, "Closing the serial connection")
      self._bus.close()



# coding:utf-8

from typing import Optional, NoReturn, Any
from importlib import import_module
import webbrowser
from pkg_resources import resource_string, resource_filename
from numpy import frombuffer, uint8


# Quick access to documentation
def docs():
  """Opens the online documentation of Crappy.

  It opens the latest version, and of course requires an internet access.

  .. versionadded:: 1.5.5
  .. versionchanged:: 2.0.0 renamed from doc to docs
  """

  webbrowser.open('https://crappy.readthedocs.io/en/latest/')


class OptionalModule:
  """Placeholder for optional dependencies that are not installed.

  Will display a message and raise an error when trying to use them.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               module_name: str,
               message: Optional[str] = None,
               lazy_import: bool = False) -> None:
    """Sets the arguments.

    Args:
      module_name: The name of the module s a :obj:`str`, preferably the one
        invoked with `pip install`.
      message: Optionally, the message to display in case the module is
        missing (as a :obj:`str`). If not provided, a generic message will be
        displayed.
      lazy_import: If :obj:`True`, the module won't be imported directly even
        if it is installed. Instead, it will be imported only when necessary.
        Allows reducing the import time, especially on Window.
    
    .. versionadded:: 2.0.0 *lazy_import* argument
    """

    self._name = module_name

    # Setting a default message if none was provided
    if message is not None:
      self._msg = message
    else:
      self._msg = f"The module {self._name} is necessary to use this " \
                  f"functionality. Please install it and try again !"

    self._lazy = lazy_import
    self._module = None

  def __getattr__(self, attr: str) -> Any:
    """Method normally raising an exception indicating that the module is
    missing.

    In case the ``lazy_import`` argument was set to :obj:`True`, still tries to
    get the desired attribute and raises the exception only if the module is
    missing.
    """

    # The module has to be imported only when called because it's too heavy
    if self._lazy:
      raise_ = False

      # Trying to import the module
      if self._module is None:
        try:
          self._module = import_module(self._name)
        except (ImportError, ModuleNotFoundError):
          raise_ = True

      # The module could be imported, returning the desired attribute
      if not raise_:
        return getattr(self._module, attr)

    # The module could not be imported
    raise RuntimeError(f"Missing module: {self._name}\n{self._msg}")

  def __call__(self, *_, **__) -> NoReturn:
    """Method raising an exception indicating that the module is missing."""

    raise RuntimeError(f"Missing module: {self._name}\n{self._msg}")


# Data aliases
class resources:
  """This class defines aliases for quick access to the resources in the
  `tool/data/` folder.

  These aliases are then used in the examples provided on the GitHub
  repository, but could also be used in custom user scripts.

  .. versionadded:: 1.5.3
  """

  try:
    # Defining aliases to the images themselves
    from cv2 import imdecode
    speckle = imdecode(frombuffer(resource_string('crappy',
                                                  'tool/data/speckle.png'),
                                  uint8), flags=0)

    ve_markers = imdecode(frombuffer(
      resource_string('crappy', 'tool/data/ve_markers.tif'), uint8), flags=0)

    pad = imdecode(frombuffer(resource_string('crappy', 'tool/data/pad.png'),
                              uint8), flags=0)

  # In case the module opencv-python is missing
  except (ModuleNotFoundError, ImportError):
    speckle = OptionalModule('opencv-python')
    ve_markers = OptionalModule('opencv-python')
    pad = OptionalModule('opencv-python')

  # Also getting the paths to the images
  paths = {'pad': resource_filename('crappy', 'tool/data/pad.png'),
           'speckle': resource_filename('crappy', 'tool/data/speckle.png'),
           've_markers': resource_filename('crappy',
                                           'tool/data/ve_markers.tif')}


class LinkDataError(ValueError):
  """Exception raised when trying to send a wrong data type through a
  :class:`~crappy.links.Link`."""


class StartTimeout(TimeoutError):
  """Exception raised when the start event takes too long to be set."""


class PrepareError(IOError):
  """Error raised in a :class:`~crappy.blocks.Block` when waiting for all
  Blocks to be ready but another Block fails to prepare."""


class CameraConfigError(RuntimeError):
  """Error raised by a :class:`~crappy.tool.camera_config.CameraConfig` window
  when encountering an exception."""


class CameraPrepareError(RuntimeError):
  """Error raised by a :class:`~crappy.blocks.Camera` when one of its children
  processes crashes while preparing."""


class CameraRuntimeError(RuntimeError):
  """Error raised by a :class:`~crappy.blocks.Camera` when one of its children
  processes crashes while running."""


class T0NotSetError(ValueError):
  """Exception raised when requesting the t0 value when it is not set."""


class DefinitionError(NameError):
    """Exception raised when trying to define an object with the same name as
    an already-defined one."""

    def __init__(self, msg: str) -> None:
      """Sets the msg attribute.

      Args:
        msg: The message to display along with the exception.
      """

      super().__init__()
      self._msg = msg

    def __str__(self) -> str:
      return self._msg


class GeneratorStop(Exception):
  """Exception raised when a :class:`~crappy.blocks.Generator` Block reaches
  the end of its :class:`~crappy.blocks.generator_path.Path`."""


class ReaderStop(Exception):
  """Exception raised when a :class:`~crappy.camera.FileReader` Camera has
  exhausted all the images to read."""


class CrappyFail(Exception):
  """Exception raised when an unexpected Exception is caught in Crappy. Also
  raised when all the Blocks do not terminate gracefully."""

  def __init__(self) -> None:
    """Only defined here to set the message to display when raised."""

    super().__init__("This Exception means that an error occurred while "
                     "running Crappy. Check the Traceback in the terminal for "
                     "more information")



# coding: utf-8

import numpy as np
from typing import Optional, Tuple, List, Union
from pathlib import Path
import logging
import logging.handlers

from .camera_process import CameraProcess
from ...tool.image_processing import GPUCorrelTool
from ...tool.camera_config import SpotsBoxes
from ..._global import OptionalModule

try:
  import pycuda.tools
  import pycuda.driver
except (ModuleNotFoundError, ImportError):
  pycuda = OptionalModule("pycuda")


class GPUVEProcess(CameraProcess):
  """This :class:`~crappy.blocks.camera_processes.CameraProcess` can perform
  GPU-accelerated video-extensometry by tracking patches using digital image
  correlation. It then returns the position of the tracked patches.

  It is used by the :class:`~crappy.blocks.GPUVE` Block to parallelize the
  image processing and the image acquisition. It delegates most of the
  computation to the :class:`~crappy.tool.image_processing.GPUCorrelTool`. It
  is from this class that the output values are sent to the downstream Blocks,
  and that the :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` are
  sent to the :class:`~crappy.blocks.camera_processes.Displayer` CameraProcess
  for display.

  .. versionadded:: 2.0.0
  """

  def __init__(self,
               patches: List[Tuple[int, int, int, int]],
               verbose: int = 0,
               kernel_file: Optional[Union[str, Path]] = None,
               iterations: int = 4,
               img_ref: Optional[np.ndarray] = None,
               mul: float = 3) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      patches: A :obj:`list` containing the coordinates of the patches to
        track, as a :obj:`tuple` for each patch. Each tuple should contain
        exactly `4` elements, giving in pixels the `y` origin, `x` origin,
        height and width of the patch.
      verbose: The verbose level as an integer, between `0` and `3`. At level
        `0` no information is displayed, and at level `3` so much information
        is displayed that it slows the code down. This argument is passed to
        the :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used
        in this class.
      kernel_file: The path to the file containing the kernels to use for the
        correlation. Can be a :obj:`pathlib.Path` object or a :obj:`str`. If
        not provided, the default :ref:`GPU Kernels` are used. This argument is
        passed to the :class:`~crappy.tool.image_processing.GPUCorrelTool` and
        not used in this class.
      iterations: The maximum number of iterations to run before returning the
        results. The results may be returned before if the residuals start
        increasing. This argument is passed to the
        :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used in
        this class.
      img_ref: A reference image for the correlation, given as a 2D
        :obj:`numpy.array` with `dtype` :obj:`numpy.float32`. If not given, the
        first acquired frame will be used as the reference image instead. This
        argument is passed to the
        :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used in
        this class.
      mul: The scalar by which the direction will be multiplied before being
        added to the solution. If it's too high, the convergence will be fast
        but there's a risk to go past the solution and to diverge. If it's too
        low, the convergence will be slower and require more iterations. `3`
        was found to be an acceptable value in most cases, but it is
        recommended to tune this value for each application so that the
        convergence is neither too slow nor too fast. This argument is passed
        to the :class:`~crappy.tool.image_processing.GPUCorrelTool` and not
        used in this class.
    """

    super().__init__()

    # Making a CUDA context common to all the patches
    pycuda.driver.init()
    self._context = pycuda.tools.make_default_context()
    
    # Arguments to pass to the GPUCorrelTools
    self._verbose = verbose
    self._kernel_file = kernel_file
    self._iterations = iterations
    self._ref_img = img_ref
    self._mul = mul

    # Other attributes
    self._correls: Optional[List[GPUCorrelTool]] = None
    self._patches = patches
    self._img_ref = img_ref

    self._spots = SpotsBoxes()
    self._spots.set_spots(patches)

    self._img0_set = img_ref is not None

  def init(self) -> None:
    """Initializes the GPUCorrelTool instances, and set their reference image
    if a ``img_ref`` argument was provided."""

    # Instantiating the GPUCorrelTool instances
    self.log(logging.INFO, "Instantiating the GPUCorrel tool instances")
    self._correls = [GPUCorrelTool(logger_name=self.name,
                                   context=self._context,
                                   verbose=self._verbose,
                                   levels=1,
                                   resampling_factor=2,
                                   kernel_file=self._kernel_file,
                                   iterations=self._iterations,
                                   fields=['x', 'y'],
                                   ref_img=self._img_ref,
                                   mask=None,
                                   mul=self._mul) for _ in self._patches]

    # We can already set the sizes of the images as they are already known
    self.log(logging.INFO, "Setting the sizes of the patches")
    for correl, (_, __, h, w) in zip(self._correls, self._patches):
      correl.set_img_size((h, w))

    # Setting the reference image if it was given as an argument
    if self._img_ref is not None:
      self.log(logging.INFO, "Initializing the GPUCorrel tool instances "
                             "with the given reference image and preparing "
                             "them")
      for correl, (oy, ox, h, w) in zip(self._correls, self._patches):
        correl.set_orig(
          self._img_ref[oy:oy + h, ox:ox + w].astype(np.float32))
        correl.prepare()

  def loop(self) -> None:
    """This method grabs the latest frame and gives it for processing to the
    several instances of :class:`~crappy.tool.image_processing.GPUCorrelTool`.
    Then sends the displacement data to the downstream Blocks.

    If there's no new frame grabbed, doesn't do anything. On the first acquired
    frame, does not process it but initializes the instances of GPUCorrelTool
    with it instead if no reference image was given as argument. Also sends the
    patches for display to the
    :class:`~crappy.blocks.camera_processes.Displayer` CameraProcess.
    """

    # Setting the reference image with the first received frame if it was not
    # given as an argument
    if not self._img0_set:
      self.log(logging.INFO, "Setting the reference image")
      for correl, (oy, ox, h, w) in zip(self._correls, self._patches):
        correl.set_orig(self.img[oy:oy + h, ox:ox + w].astype(np.float32))
        correl.prepare()
      self._img0_set = True
      return

    # Performing the image correlation
    self.log(logging.DEBUG, "Processing the received image")
    data = [self.metadata['t(s)'], self.metadata]
    for correl, (oy, ox, h, w) in zip(self._correls, self._patches):
      data.extend(correl.get_disp(
          self.img[oy:oy + h, ox:ox + w].astype(np.float32)).tolist())

    # Sending the data to downstream Blocks
    self.send(data)

    # Sending the patches to the Displayer for display
    self.send_to_draw(self._spots)

  def finish(self) -> None:
    """Performs cleanup on the several
    :class:`~crappy.tool.image_processing.GPUCorrelTool` used."""

    if self._correls is not None:
      self.log(logging.INFO, "Cleaning up the GPUCorrel instances")
      for correl in self._correls:
        correl.clean()



# coding: utf-8

from typing import Optional
import logging

from .meta_block import Block


class LinkReader(Block):
  """This Block reads and displays the data it receives.
  
  It can display data received from any number of :class:`~crappy.links.Link`.
  For each new received data point, a message is displayed in the console with
  the received values.
  
  This Block is the most basic way of displaying data in Crappy. The 
  :class:`~crappy.blocks.Dashboard` Block can be used for a nicer layout, and 
  the :class:`~crappy.blocks.Grapher` Block should be used for plotting data in
  a persistent way.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Reader to LinkReader
  """

  _index = 0

  def __init__(self,
               name: Optional[str] = None,
               freq: Optional[float] = 50,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      name: If set, will be displayed to identify the LinkReader. Otherwise, 
        the block is automatically named based on the number of its instances 
        already running.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      display_freq: if :obj:`True`, displays the looping frequency of the 
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.

    .. versionchanged:: 1.5.5 renamed *name* argument to *reader_name*
    .. versionadded:: 1.5.10 *freq* and *verbose* arguments
    .. versionchanged:: 1.5.5 renamed *reader_name* argument to *name*
    .. versionadded 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    self._reader_name = name if name is not None else f'LinkReader ' \
                                                      f'{self._get_index()}'

  def __new__(cls, *args, **kwargs):
    """Increments the instance counter when instantiating a new LinkReader."""

    cls._index += 1
    return super().__new__(cls)

  @classmethod
  def _get_index(cls) -> int:
    """Returns the current number of instantiates Links, as an :obj:`int`."""

    return cls._index

  def loop(self) -> None:
    """Flushes the incoming :class:`~crappy.links.Link` and displays their
    data."""

    data = self.recv_all_data_raw()
    for link_data in data:
      for dic in (dict(i) for i in
                  zip(*([(key, value) for value in values]
                        for key, values in link_data.items()))):
        self.log(logging.INFO, f'{self._reader_name} got: {dic}')



# coding: utf-8

from typing import Optional
import logging

from .meta_inout import InOut
from .._global import OptionalModule

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")


class GPIOPWM(InOut):
  """This class can drive a PWM output on a Raspberry Pi.

  It allows controlling the duty cycle, the frequency, or both at the same
  time. When controlling both, the duty cycle should be first and the frequency
  second in the given commands.

  Warning:
    Only works on Raspberry Pi !
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Gpio_pwm to GPIOPWM
  """

  def __init__(self,
               pin_out: int,
               duty_cycle: Optional[float] = None,
               frequency: Optional[float] = None) -> None:
    """Checks the validity of the arguments.

    Args:
      pin_out: The index of GPIO pin to drive in BCM convention, as an
        :obj:`int`.
      duty_cycle: If provided (as a :obj:`float`, in percent), sets a fixed
        duty cycle for the entire test. Only the frequency can then be tuned.
        If not provided, the duty cycle can be set as a command. The duty cycle
        will also be set to `0%` until a first value is received.
      frequency: If provided (as a :obj:`float`, in Hz), sets a fixed PWM
        frequency for the entire test. Only the duty cycle can then be tuned.
        If not provided, the frequency can be set as a command. The frequency
        will also be set to `10kHz` until a first value is received. Note that
        the frequency inputs are clamped between `10Hz` and `1MHz`.

    Note:
      Several values can be passed at once as a command. If both ``duty_cycle``
      and ``frequency`` are provided, all the values are ignored. If only
      ``frequency`` is provided, the first command value sets the duty cycle
      and any other value is ignored. Same goes if only ``duty_cycle`` is
      provided. If none of the two arguments are provided, the first command
      value should set the duty cycle and the second command value should set
      the frequency.

    Note:
      On the Raspberry Pi 4, only the GPIO pins `12`, `13`, `18` and `19`
      support hardware PWM. Trying to get a PWM output from other pins might
      work but may decrease the available frequency range.
    """

    self._pwm = None

    super().__init__()

    if pin_out not in range(2, 28):
      raise ValueError("pin_out should be an integer between 2 and 28")
    else:
      self._pin_out = pin_out

    if frequency is not None:
      if not 10 <= frequency < 1000000:
        raise ValueError("frequency should be between 100Hz and 1MHz")
    self._frequency = frequency

    if duty_cycle is not None:
      if not 0 <= duty_cycle <= 100:
        raise ValueError("Duty cycle should be positive and not exceed 100%")
    self._duty_cycle = duty_cycle

  def open(self) -> None:
    """Sets the GPIOs and starts the PWM."""

    # Setting the GPIOs
    self.log(logging.INFO, "Setting up the GPIOs")
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(self._pin_out, GPIO.OUT)

    # Setting to user frequency if provided, or else to 10kHz
    self.log(logging.INFO, "Setting up the PWM")
    if self._frequency is not None:
      self._pwm = GPIO.PWM(self._pin_out, self._frequency)
    else:
      self._pwm = GPIO.PWM(self._pin_out, 10000)

    # Setting to user duty cycle if provided, or else to 0%
    self.log(logging.INFO, "Starting the PWM")
    if self._duty_cycle is not None:
      self._pwm.start(self._duty_cycle)
    else:
      self._pwm.start(0)

  def set_cmd(self, *cmd: float) -> None:
    """Modifies the PWM frequency and/or duty cycle.

    Args:
      *cmd: Values of duty cycle and/or frequency to set.
    """

    # If both frequency and duty cycle are fixed by the user, nothing to do
    if self._duty_cycle is not None and self._frequency is not None:
      return

    # If only frequency is fixed, setting the duty cycle
    elif self._frequency is not None:
      dc = min(100., max(0., cmd[0]))
      self._pwm.ChangeDutyCycle(dc)

    # If only the duty cycle is fixed, setting the frequency
    elif self._duty_cycle is not None:
      freq = min(1000000., max(10., cmd[0]))
      self._pwm.ChangeFrequency(freq)

    # If neither duty cycle nor frequency are fixed, setting both
    else:
      dc = min(100., max(0., cmd[0]))
      freq = min(1000000., max(10., cmd[1]))
      self._pwm.ChangeFrequency(freq)
      self._pwm.ChangeDutyCycle(dc)

  def close(self) -> None:
    """Stops the PWM and releases the GPIOs."""

    if self._pwm is not None:
      self.log(logging.INFO, "Stopping the PWM")
      self._pwm.stop()

    self.log(logging.INFO, "Cleaning up the GPIOs")
    GPIO.cleanup()



# coding: utf-8

from typing import Optional, Callable, Union
import logging

from .camera_setting import CameraSetting

NbrType = Union[int, float]


class CameraScaleSetting(CameraSetting):
  """Camera setting that can take any value between a lower and an upper
  boundary.

  It is a child of
  :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting`.

  This class can handle settings that should only take :obj:`int` values as
  well as settings that can take :obj:`float` value. The type used is
  :obj:`int` is both of the given lowest or highest values are :obj:`int`,
  otherwise :obj:`float` is used.
  
  .. versionadded:: 1.5.10
  .. versionchanged:: 2.0.0
     renamed from Camera_scale_setting to CameraScaleSetting
  """

  def __init__(self,
               name: str,
               lowest: NbrType,
               highest: NbrType,
               getter: Optional[Callable[[], NbrType]] = None,
               setter: Optional[Callable[[NbrType], None]] = None,
               default: Optional[NbrType] = None,
               step: Optional[NbrType] = None) -> None:
    """Sets the attributes.

    Args:
      name: The name of the setting, that will be displayed in the GUI.
      lowest: The lower boundary for the setting values.
      highest: The upper boundary for the setting values.
      getter: The method for getting the current value of the setting.
      setter: The method for setting the current value of the setting.
      default: The default value to assign to the setting.
      step: The step value for the variation of the setting values.
      
    .. versionadded:: 2.0.0 add *step* argument
    """

    self.lowest = lowest
    self.highest = highest
    self.step = step
    self.type = int if isinstance(self.lowest + self.highest, int) else float

    if default is None:
      default = self.type((self.lowest + self.highest) / 2)
    else:
      default = self.type(default)

    super().__init__(name, getter, setter, default)

    self._check_value()

  @property
  def value(self) -> NbrType:
    """Returns the current value of the setting, by calling the getter if one
    was provided or else by returning the stored value.

    When the getter is called, calls the setter if one was provided and updates
    the sored value. After calling the setter, checks that the value was set
    by calling the getter and displays a warning message if the target and
    actual values don't match.
    """

    if self._getter is not None:
      return self.type(min(max(self._getter(), self.lowest), self.highest))
    else:
      return self.type(self._value_no_getter)

  @value.setter
  def value(self, val: NbrType) -> None:
    val = min(max(val, self.lowest), self.highest)
    self.log(logging.DEBUG, f"Setting the setting {self.name} to {val}")

    self._value_no_getter = self.type(val)
    if self._setter is not None:
      self._setter(self.type(val))

    if self.value != val:
      # Double-checking, got strange behavior sometimes probably because of
      # delays in lower level APIs
      if self.value == val:
        return
      self.log(logging.WARNING, f"Could not set {self.name} to {val}, the "
                                f"value is {self.value} !")

  def reload(self,
             lowest: NbrType,
             highest: NbrType,
             value: NbrType,
             default: Optional[NbrType] = None,
             step: Optional[NbrType] = None) -> None:
    """Allows modifying the limits and the step of the scale bar once it is 
    already instantiated.
    
    .. versionadded:: 2.0.0
    """

    self.log(logging.DEBUG, f"Reloading the setting {self.name}")

    # Updating the lowest, highest, step and default values
    self.lowest = lowest
    self.highest = highest
    self.step = step

    if default is not None:
      self.default = self.type(default)
    else:
      self.default = self.type((lowest + highest) / 2)

    self._check_value()

    # Updating the slider limits and the setting value
    if self.tk_obj is not None:
      self.tk_obj.configure(to=self.highest,
                            from_=self.lowest,
                            resolution=self.step)
    if self.tk_var is not None:
      self.tk_var.set(value)

  def _check_value(self) -> None:
    """Checks if the step value is compatible with the limit values and
    types of the scale settings.
    
    .. versionadded:: 2.0.0
    """

    if self.step is not None:
      if self.type == int and isinstance(self.step, float):
        self.step = max(int(self.step), 1)
        self.log(logging.WARNING, f"Could not set {self.name} step "
                                  f"(lowest: int, step: float), "
                                  f"the step is now {self.step} !")
      if self.highest > self.lowest and self.step > self.highest - self.lowest:
        self.step = 1 if self.type == int else (self.highest -
                                                self.lowest) / 1000
        self.log(logging.WARNING, f"Could not set {self.name} step, "
                                  f"the step is now {self.step} !")
      if self.type == int and (self.highest - self.lowest) % self.step:
          self.highest -= (self.highest - self.lowest) % self.step
          self.log(logging.WARNING, f"Could not set {self.name} highest "
                                    f"with this step {self.step},"
                                    f" the highest is now {self.highest} !")
    else:
      self.step = 1 if self.type == int else (self.highest -
                                              self.lowest) / 1000



# coding: utf-8

from time import time
from typing import Optional, Tuple
import logging

from .meta_block import Block


class PID(Block):
  """This Block implements a basic PID corrector.
  
  It takes a target value and a measured value as an input, and outputs a
  command value for reaching the target on the controlled system. The command
  is updated at each loop. The given `P`, `I` and `D` gains can be adjusted at
  runtime, by sending the new desired values. It is also possible to set 
  boundaries to the output command, or to the integral corrector only.
  
  This Block is generally used in combination with a 
  :class:`~crappy.blocks.Generator` Block for generating the target value, and
  a :class:`~crappy.blocks.Machine` or :class:`~crappy.blocks.IOBlock` Block
  driving a physical system. This latter Block takes the command as input, and
  returns to the PID the measured value to be compared to the target.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               kp: float,
               ki: float = 0,
               kd: float = 0,
               out_max: float = float('inf'),
               out_min: float = -float('inf'),
               setpoint_label: str = 'cmd',
               input_label: str = 'V',
               time_label: str = 't(s)',
               kp_label: str = 'kp',
               ki_label: str = 'ki',
               kd_label: str = 'kd',
               labels: Optional[Tuple[str, str]] = None,
               reverse: bool = False,
               i_limit: Tuple[Optional[float], Optional[float]] = (None, None),
               send_terms: bool = False,
               freq: Optional[float] = 500,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      kp: The initial `P` gain. It can be tuned while running by sending the
        new value over the given ``kp_label``. No matter if a positive or a
        negative value is given, the definitive sign will be set by the
        ``reverse`` argument.
      ki: The initial `I` gain. It can be tuned while running by sending the
        new value over the given ``ki_label``. No matter if a positive or a
        negative value is given, the definitive sign will be set by the
        ``reverse`` argument.
      kd: The initial `D` gain. It can be tuned while running by sending the
        new value over the given ``kd_label``. No matter if a positive or a
        negative value is given, the definitive sign will be set by the
        ``reverse`` argument.
      out_max: Ensures the output is always inferior to this value.
      out_min: Ensures the output is always superior to this value.
      setpoint_label: The label carrying the setpoint value.
      input_label: The label carrying the reading of the actual value, to be
        compared with the setpoint.
      time_label: The label carrying the time information in the incoming
        Links.
      kp_label: The label to use for changing the `P` gain on the fly. If a
        value is received over this label, it will overwrite the one given in
        the ``kp`` argument.
      ki_label: The label to use for changing the `I` gain on the fly. If a
        value is received over this label, it will overwrite the one given in
        the ``ki`` argument.
      kd_label: The label to use for changing the `D` gain on the fly. If a
        value is received over this label, it will overwrite the one given in
        the ``kd`` argument.
      labels: The two labels that will be sent to downstream Blocks. The first
        one is the time label, the second one is the output of the PID. If this
        argument is not given, they default to ``'t(s)'`` and ``'pid'``.
      reverse: If :obj:`True`, reverses the action of the PID.
      i_limit: A :obj:`tuple` containing respectively the lower and upper
        boundaries for the `I` term.
      send_terms: If :obj:`True`, returns the weight of each term in the output
        value. It adds ``'p_term', 'i_term', 'd_term'`` to the output labels.
        This is particularly useful to tweak the gains.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionadded:: 1.5.10 *verbose* argument
    .. versionadded 2.0.0 
       *debug*, *kp_label*, *kd_label* and *ki_label* arguments
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    .. versionchanged:: 2.0.0
       renamed *target_label* argument to *setpoint_label*
    """

    # Attributes of the parent class
    super().__init__()
    self.niceness = -10
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug
    self.labels = ['t(s)', 'pid'] if labels is None else list(labels)
    if send_terms:
      self.labels.extend(['p_term', 'i_term', 'd_term'])

    # Setting the gains
    sign = -1 if reverse else 1
    self._kp = sign * abs(kp)
    self._ki = sign * abs(ki)
    self._kd = sign * abs(kd)

    # Setting the limits
    self._out_max = out_max
    self._out_min = out_min
    self._i_min, self._i_max = i_limit

    # Setting the labels
    self._target_label = setpoint_label
    self._input_label = input_label
    self._time_label = time_label
    self._kp_label = kp_label
    self._ki_label = ki_label
    self._kd_label = kd_label

    self._send_terms = send_terms
    self._reverse = reverse

    # Setting the variables
    self._setpoint: Optional[float] = None
    self._last_input: Optional[float] = None
    self._prev_t: float = 0.
    self._i_term: float = 0.

  def loop(self) -> None:
    """Receives the latest target and input values, calculates the P, I and D
    terms and sends the output to the downstream Blocks."""

    # Looping in a non-blocking way
    data = self.recv_last_data(fill_missing=False)

    # Updating the gains if provided
    if self._kp_label in data:
      kp = data[self._kp_label]
      self._kp = -abs(kp) if self._reverse else kp
    if self._ki_label in data:
      ki = data[self._ki_label]
      self._ki = -abs(ki) if self._reverse else ki
    if self._kd_label in data:
      kd = data[self._kd_label]
      self._kd = -abs(kd) if self._reverse else kd

    # Updating the target value if provided
    if self._target_label in data:
      self._setpoint = data[self._target_label]
      self.log(logging.DEBUG, f"Updated target value to {self._setpoint}")

    # Checking if a new input was received
    if self._time_label in data and self._input_label in data:
      input_ = data[self._input_label]
      t = data[self._time_label]
      self.log(logging.DEBUG, f"Updated input value to {input_} at time {t}")

      # For the first loops, setting the target to the first input by default
      if self._setpoint is None:
        self._setpoint = input_

      # For the first loops, initializing the input history
      if self._last_input is None:
        self._last_input = input_

    # No new input was received
    else:
      return

    delta_t = t - self._prev_t
    error = self._setpoint - input_
    d_input = input_ - self._last_input

    # Calculating the three PID terms
    p_term = self._kp * error
    self._i_term += self._ki * error * delta_t
    d_term = - self._kd * d_input / delta_t if delta_t > 0 else 0

    self._prev_t = t
    self._last_input = input_

    # Clamping the i term if required
    if self._i_min is not None:
      self._i_term = max(self._i_min, self._i_term)
    if self._i_max is not None:
      self._i_term = min(self._i_max, self._i_term)

    # Clamping the output if required
    out = p_term + self._i_term + d_term
    out = min(self._out_max, max(self._out_min, out))

    # Sending the values to the downstream blocks
    if self._send_terms:
      self.send([time() - self.t0, out, p_term, self._i_term, d_term])
    else:
      self.send([time() - self.t0, out])



# coding: utf-8

from time import time, sleep
from typing import Tuple, Union, Optional
import numpy as np
from pathlib import Path
from re import fullmatch
import logging

from .meta_camera import Camera
from .._global import OptionalModule, ReaderStop

try:
  import SimpleITK as Sitk
except (ModuleNotFoundError, ImportError):
  Sitk = OptionalModule("SimpleITK")
try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class FileReader(Camera):
  """This Camera class reads existing images from a given folder, in the same
  order in which they were acquired.

  The name of the images to read must follow the following pattern :
  ``<frame_nr>_<frame_seconds>.<frame_subseconds>.<file_extension>``. This
  pattern is the same as used for recording images with the
  :class:`~crappy.blocks.Camera` of Crappy, so images recorded via Crappy are
  readily readable and don't need to be re-named.

  This class tries to read the images at the same framerate as they were
  recorded, although the control of the framerate is not so precise. It might
  be that the images cannot be read fast enough to match the original
  framerate, in which case the images are read as fast as possible and the
  delay keeps growing.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 1.5.10 renamed from Streamer to File_reader
  .. versionchanged:: 2.0.0 renamed from File_reader to FileReader
  """

  def __init__(self) -> None:
    """Initializes the parent class and sets a few attributes."""

    super().__init__()

    # These attributes will come in use later on
    self._images = None
    self._stop_at_end = True
    self._backend = None
    self._t0 = None
    self._stopped = False

  def open(self,
           reader_folder: Union[Path, str],
           reader_backend: Optional[str] = None,
           stop_at_end: bool = True) -> None:
    """Sets the reader backend and retrieves the images to read, sorted by
    their timestamp.

    Args:
      reader_folder: The path to the folder containing the images to read.
      reader_backend: The backend to use for reding the images. Should be one
        of :
        ::

          'sitk' or 'cv2'

        If not given, SimpleITK is preferred over OpenCV if available.
      stop_at_end: If :obj:`True` (the default), stops the Crappy script once
        the available images are all exhausted. Otherwise, simply remains idle
        while waiting for the test to finish.
    
    .. versionchanged:: 1.5.10 renamed *path* argument to *reader_folder*
    .. versionadded:: 1.5.10 *reader_backend* and *stop_at_end* arguments
    .. deprecated:: 1.5.10 *pattern*, *start_delay* and *modifier* arguments
    """

    # Selecting an  available backend between first sitk and then cv2
    if reader_backend is None:
      if not isinstance(Sitk, OptionalModule):
        self._backend = 'sitk'
      elif not isinstance(cv2, OptionalModule):
        self._backend = 'cv2'
      else:
        raise ModuleNotFoundError("Neither SimpleITK nor opencv-python could "
                                  "be imported, no backend found for reading "
                                  "the images")
    # Setting the backend requested by the user
    elif reader_backend in ('sitk', 'cv2'):
      self._backend = reader_backend
    else:
      raise ValueError("The backend argument should be either 'sitk' or "
                       "'cv2' !")

    self._stop_at_end = stop_at_end

    # Making sure that the given folder is valid
    folder = Path(reader_folder)
    if not folder.exists() or not folder.is_dir():
      raise FileNotFoundError(f"The {folder} folder does not exist or is not "
                              f"a folder !")

    # Retrieving all the images in the given folder that match the name pattern
    images = (path for path in folder.glob('*') if
              fullmatch(r'\d+_\d+\.\d+\..+\Z', path.name) is not None)
    # Sorting the images by timestamp
    images = sorted(images,
                    key=lambda p: float(fullmatch(r'\d+_(\d+\.\d+)',
                                                  p.stem).group(1)))

    # In case no matching image was found
    if not images:
      raise FileNotFoundError(f"Could not find matching images in the {folder}"
                              f" folder !\nPlease specify a valid folder or "
                              f"put the image names in the right format.")

    self.log(logging.INFO, f"Detected {len(images)} images in the {folder} "
                           f"folder")

    # The images are stored as an iterator
    self._images = iter(images)

  def get_image(self) -> Optional[Tuple[float, np.ndarray]]:
    """Reads the next image in the image folder, and returns it at the right
    time so that the achieved framerate matches the original framerate.

    If the original framerate cannot be achieved, just reads the image as fast
    as possible.

    By default, stops the test when there's no image left to read. If specified
    otherwise, just remains idle until the test ends.
    """

    # Setting the approximate start time (potentially not well synced with the
    # actual t0 of Crappy's blocks)
    if self._t0 is None:
      self._t0 = time()

    if self._stopped:
      sleep(0.1)
      return

    try:
      # Getting the next image to read and its timestamp
      img_path = next(self._images)
      timestamp = float(fullmatch(r'\d+_(\d+\.\d+)', img_path.stem).group(1))

      self.log(logging.DEBUG, f"Reading image {img_path} with timestamp "
                              f"{timestamp}")

      # Reading the image data with the chosen backend
      if self._backend == 'sitk':
        img = Sitk.GetArrayFromImage(Sitk.ReadImage(img_path))
      else:
        img = cv2.imread(str(img_path), 0)

      # Delaying the return of the image if we're ahead of time
      t = time()
      if t - self._t0 < timestamp:
        sleep(timestamp - (t - self._t0))

      return t, img

    # Raised when there's no more image to read
    except StopIteration:
      # Default behavior, stop the test
      if self._stop_at_end:
        raise ReaderStop
      else:
        # Otherwise, nothing more gets done but the test goes on
        self._stopped = True
        self.log(logging.WARNING, "Exhausted all the images to read for the "
                                  "FileReader camera, staying idle until the "
                                  "script ends")



# coding: utf-8

from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, List, Tuple

from .box import Box


@dataclass
class SpotsBoxes:
  """This class stores up to four instances of
  :class:`~crappy.tool.camera_config.config_tools.Box`, defining the bounding
  boxes of the spots for :class:`~crappy.blocks.VideoExtenso` or the patches
  for :class:`~crappy.blocks.DICVE`.

  It can also instantiate the Box objects by parsing a list of tuples
  containing enough information.

  .. versionadded:: 2.0.0
  """

  spot_1: Optional[Box] = None
  spot_2: Optional[Box] = None
  spot_3: Optional[Box] = None
  spot_4: Optional[Box] = None

  x_l0: Optional[float] = None
  y_l0: Optional[float] = None

  _index = -1

  def __getitem__(self, i: int) -> Optional[Box]:
    if i == 0:
      return self.spot_1
    elif i == 1:
      return self.spot_2
    elif i == 2:
      return self.spot_3
    elif i == 3:
      return self.spot_4
    else:
      raise IndexError

  def __setitem__(self, i: int, value: Optional[Box]) -> None:
    if i == 0:
      self.spot_1 = value
    elif i == 1:
      self.spot_2 = value
    elif i == 2:
      self.spot_3 = value
    elif i == 3:
      self.spot_4 = value
    else:
      raise IndexError

  def __iter__(self) -> SpotsBoxes:
    self._index = -1
    return self

  def __next__(self) -> Box:
    self._index += 1
    try:
      return self[self._index]
    except IndexError:
      raise StopIteration

  def __len__(self) -> int:
    return len([spot for spot in self if spot is not None])

  def set_spots(self,
                spots: List[Tuple[int, int, int, int]]) -> None:
    """Parses a list of tuples and instantiates the corresponding
    :class:`~crappy.tool.camera_config.config_tools.Box` objects."""

    for i, spot in enumerate(spots):
      self[i] = Box(x_start=spot[1], x_end=spot[1] + spot[3],
                    y_start=spot[0], y_end=spot[0] + spot[2])

  def save_length(self) -> None:
    """Setting the :attr:`x_l0` and :attr:`y_l0` attributes based on the
    positions of the centroids.

    If only one spot is detected, setting the initial lengths to 0.
    """

    # Calculating the centroids of the spots if not already known
    for spot in self:
      if spot is not None and spot.x_centroid is None:
        min_x, max_x, min_y, max_y = spot.sorted()
        spot.x_centroid = min_x + (max_x - min_x) / 2
        spot.y_centroid = min_y + (max_y - min_y) / 2

    # Simply taking the distance between the extrema as the initial length
    if len(self) > 1:
      x_centers = [spot.x_centroid for spot in self if spot is not None]
      y_centers = [spot.y_centroid for spot in self if spot is not None]
      self.x_l0 = max(x_centers) - min(x_centers)
      self.y_l0 = max(y_centers) - min(y_centers)

    # If only one spot detected, setting the initial lengths to 0
    else:
      self.x_l0 = 0
      self.y_l0 = 0

  def empty(self) -> bool:
    """Returns :obj:`True` if all spots are :obj:`None`, else :obj:`False`."""

    return all(spot is None for spot in self)

  def reset(self) -> None:
    """Resets the boxes to :obj:`None`."""

    for i in range(4):
      self[i] = None



# coding: utf-8

from subprocess import check_output
from threading import Thread, RLock
from time import sleep
from typing import Union, Dict, Optional
import logging

from .meta_actuator import Actuator
from .._global import OptionalModule

try:
  import yaml
  is_installed = True
except (ModuleNotFoundError, ImportError):
  is_installed = False
try:
  getattr(yaml, 'FullLoader')
  full_loader = True
except (AttributeError, NameError):
  full_loader = False
if not is_installed:
  # If it was in the first try/except, a message would be displayed at getattr
  yaml = OptionalModule("pyyaml")

try:
  from usb import core
  from usb import util

  Tic_usb_request = {'Cmd': util.CTRL_OUT |
                     util.CTRL_TYPE_VENDOR |
                     util.CTRL_RECIPIENT_DEVICE,
                     'Var': util.CTRL_IN |
                     util.CTRL_TYPE_VENDOR |
                     util.CTRL_RECIPIENT_DEVICE}
except (ModuleNotFoundError, ImportError):
  util = core = OptionalModule("pyusb")
  Tic_usb_request = {'Cmd': 0x40,
                     'Var': 0xC0}

Tic_vendor_id = 0x1FFB

Tic_product_id = {'T825': 0x00B3,
                  'T834': 0x00B5,
                  'T500': 0x00BD,
                  'N825': 0x00C3,
                  'T249': 0x00C9,
                  '36v4': 0x00CB}

Tic_max_allowed_current = {'T825': 3968,
                           'T834': 3456,
                           'T500': 3093,
                           'T249': 4480,
                           '36v4': 3939}

Tic_36v4_max_current = 9095

Tic_current_steps = {'T834': 32,
                     'T825': 32,
                     'T249': 40,
                     '36v4': 71.615}

Tic_step_modes = {'T825': [2 ** i for i in range(6)],
                  'T834': [2 ** i for i in range(6)],
                  'T500': [2 ** i for i in range(4)],
                  'T249': [2 ** i for i in range(6)] + ['2_100p'],
                  '36v4': [2 ** i for i in range(9)]}

Tic_step_mode = {1: 0,
                 2: 1,
                 4: 2,
                 8: 3,
                 16: 4,
                 32: 5,
                 '2_100p': 6,
                 64: 7,
                 128: 8,
                 256: 9}

Tic_cmd = {'Set_target_position': 0xE0,
           'Set_target_velocity': 0xE3,
           'Halt_and_set_position': 0xEC,
           'Halt_and_hold': 0x89,
           'Go_home': 0x97,
           'Reset_command_timeout': 0x8C,
           'Deenergize': 0x86,
           'Energize': 0x85,
           'Exit_safe_start': 0x83,
           'Enter_safe_start': 0x8F,
           'Reset': 0xB0,
           'Clear_driver_error': 0x8A,
           'Set_max_speed': 0xE6,
           'Set_starting_speed': 0xE5,
           'Set_max_accel': 0xEA,
           'Set_max_decel': 0xE9,
           'Set_step_mode': 0x94,
           'Set_current_limit': 0x91,
           'Set_decay_mode': 0x92,
           'Set_AGC_option': 0x98,
           'Get_variable': 0xA1,
           'Get_variable_and_clear_errors_occurred': 0xA2,
           'Get_setting': 0xA8,
           'Set_setting': 0x13,
           'Reinitialize': 0x10,
           'Start_bootloader': 0xFF,
           'Get_debug_data': 0x20}

# offsets/indexes
Tic_var = {'Operation_state': 0x00,
           'Misc_flags1': 0x01,
           'Error_status': 0x02,
           'Errors_occurred': 0x04,
           'Planning_mode': 0x09,
           'Target_position': 0x0A,
           'Target_velocity': 0x0E,
           'Starting_speed': 0x12,
           'Max_speed': 0x16,
           'Max_decel': 0x1A,
           'Max_accel': 0x1E,
           'Current_position': 0x22,
           'Current_velocity': 0x26,
           'Acting_target_position': 0x2A,
           'Time_since_last_step': 0x2E,
           'Device_reset': 0x32,
           'Vin_voltage': 0x33,
           'Up_time': 0x35,
           'Encoder_position': 0x39,
           'RC_pulse_width': 0x3D,
           'Analog_reading_SCL': 0x3F,
           'Analog_reading_SDA': 0x41,
           'Analog_reading_TX': 0x43,
           'Analog_reading_RX': 0x45,
           'Digital_readings': 0x47,
           'Pin_states': 0x48,
           'Step_mode': 0x49,
           'Current_limit': 0x4A,
           'Decay_mode': 0x4B,
           'Input_state': 0x4C,
           'Input_after_averaging': 0x4D,
           'Input_after_hysteresis': 0x4F,
           'Input_after_scaling': 0x51}

# indexes
Tic_settings = {'Setting_not_initialized': 0x00,
                'Control_mode': 0x01,
                'Never_sleep': 0x02,
                'Disable_safe_start': 0x03,
                'Ignore_err_line_high': 0x04,
                'Serial_baud_rate_generator': 0x05,
                'Serial_device_number': 0x07,
                'Auto_clear_driver_error': 0x08,
                'Command_timeout_low': 0x09,
                'Command_timeout_high': 0x0A,
                'Serial_CRC_enabled': 0x0B,
                'Low_vin_timeout': 0x0C,
                'Low_vin_shutoff_voltage': 0x0E,
                'Low_vin_startup_voltage': 0x10,
                'High_vin_shutoff_voltage': 0x12,
                'Vin_calibration': 0x14,
                'RC_max_pulse_period': 0x16,
                'RC_bad_signal_timeout': 0x18,
                'RC_consecutive_good_pulses': 0x1A,
                'Invert_motor_direction': 0x1B,
                'Input_error_min': 0x1C,
                'Input_error_max': 0x1E,
                'Input_scaling_degree': 0x20,
                'Input_invert': 0x21,
                'Input_min': 0x22,
                'Input_neutral_min': 0x24,
                'Input_neutral_max': 0x26,
                'Input_max': 0x28,
                'Output_min': 0x2A,
                'Input_averaging_enabled': 0x2E,
                'Input_hysteresis': 0x2F,
                'Current_limit_during_error': 0x31,
                'Output_max': 0x32,
                'Switch_polarity_map': 0x36,
                'Encoder_postscaler': 0x37,
                'SCL_config': 0x3B,
                'SDA_config': 0x3C,
                'TX_config': 0x3D,
                'RX_config': 0x3E,
                'RC_config': 0x3F,
                'Current_limit': 0x40,
                'Step_mode': 0x41,
                'Decay_mode': 0x42,
                'Starting_speed': 0x43,
                'Max_speed': 0x47,
                'Max_decel': 0x4B,
                'Max_accel': 0x4F,
                'Soft_error_response': 0x53,
                'Soft_error_position': 0x54,
                'Encoder_prescaler': 0x58,
                'Encoder_unlimited': 0x5C,
                'Kill_switch_map': 0x5D,
                'Serial_response_delay': 0x5E,
                'Limit_switch_forward_map': 0x5F,
                'Limit_switch_reverse_map': 0x60,
                'Homing_speed_towards': 0x61,
                'Homing_speed_away': 0x65,
                'Serial_device_number_high': 0x69,
                'Serial_alt_device_number': 0x6A,
                'Size': 0x5F,
                'Unrestricted_current_limit': 0x6C}

Tic_current_tables = {'T500': [0, 1, 174, 343, 495, 634, 762, 880, 990, 1092,
                               1189, 1281, 1368, 1452, 1532, 1611, 1687, 1762,
                               1835, 1909, 1982, 2056, 2131, 2207, 2285, 2366,
                               2451, 2540, 2634, 2734, 2843, 2962, 3093],
                      'T834': [index * Tic_current_steps['T834'] for index in
                               (list(range(33)) + list(range(34, 65, 2)) +
                                list(range(68, 109, 4)))],
                      'T825': [index * Tic_current_steps['T825'] for index in
                               (list(range(33)) + list(range(34, 65, 2)) +
                                list(range(68, 125, 4)))],
                      'T249': [index * Tic_current_steps['T249'] for index in
                               (list(range(33)) + list(range(34, 65, 2)) +
                                list(range(68, 113, 4)))],
                      '36v4': [index * Tic_current_steps['36v4'] for index in
                               range(128)]}

Tic_max_accel = 2147483647  # steps/s/100s
Tic_min_accel = 100  # steps/s/100s
Tic_max_speed = 500000000  # steps/10000s, i.e. a 50 kHz frequency
Tic_min_speed = 7  # steps/10000s, i.e. 1 step every 23 minutes

Tic_backends = ['ticcmd', 'USB']

Tic_pins_bit = {'SCL': 0,
                'SDA': 1,
                'TX': 2,
                'RX': 3,
                'RC': 4}
Tic_pin_modes = {'Default': 0,
                 'Kill switch': 7,
                 'Limit switch forward': 8,
                 'Limit switch reverse': 9}
Tic_pin_polarity = {'Active low': 0,
                    'Active high': 1}


class FindSerialNumber:
  """A class used for finding USB devices matching a given serial number, using
     the :meth:`usb.core.find` method."""

  def __init__(self, serial_number: str) -> None:
    self.serial_number = serial_number

  def __call__(self, device) -> bool:
    return device.serial_number == self.serial_number


class PololuTic(Actuator):
  """Class for controlling Pololu's Tic stepper motor divers.

  The PololuTic Actuator block is meant for controlling a Pololu Tic stepper
  motor driver. It can be driven in both speed and position. Several Tic models
  are supported. The length unit is the millimeter (`mm`), and time unit is the
  second (`s`).

  Important:
    **Only for Linux users:** In order to drive the Tic, the appropriate udev
    rule should be set. This is done automatically when installing `ticcmd`,
    or can be done using the `udev_rule_setter` utility in ``crappy``'s `util`
    folder. It is also possible to add it manually by running:
    ::

      $ echo "SUBSYSTEM==\\"usb\\", ATTR{idVendor}==\\"1ffb\\", \
MODE=\\"0666\\\"" | sudo tee pololu.rules > /dev/null 2>&1

    in a shell opened in ``/etc/udev/rules.d``.
    
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Pololu_tic to PololuTic
  """

  def __init__(self,
               steps_per_mm: float,
               current_limit: float,
               step_mode: Union[int, str] = 8,
               max_accel: float = 20,
               t_shutoff: float = 0,
               config_file: Optional[str] = None,
               serial_number: Optional[str] = None,
               model: Optional[str] = None,
               reset_command_timeout: bool = True,
               backend: str = 'USB',
               unrestricted_current_limit: bool = False,
               pin_function: Optional[Dict[str, str]] = None,
               pin_polarity: Optional[Dict[str, str]] = None) -> None:
    """Checks args validity, finds the right device, reads the current limit
    tables.

    Args:
      steps_per_mm: The number of full steps needed for the motor to travel
        `1 mm`. This varies according to the motor model, and can be deduced
        from the datasheet or directly measured. This value is usually between
        `50` and `500`.
      current_limit: The maximum current the motor is able to withstand, in mA.
        It is usually around `1A` for small stepper motors, and can go up to a
        few Amps. The maximum allowed ``current_limit`` value depends on the
        Tic model. The Tic 36v4 default maximum current limit can be increased
        using the ``unrestricted_current_limit`` parameter.
      step_mode: Sets the number of microsteps used for driving the motor. This
        number is always a power of `2`. The minimum number of microsteps is
        `1` (full steps), and the maximum depends on the Tic model. All models
        however support modes `1` to `8`. The speed and length conversions are
        managed automatically so that changing the step mode doesn't affect the
        motor behaviour.
      max_accel: The maximum allowed acceleration for the motor, in `mm/s²`.
        When asked to reach a given speed or position, the motor accelerates at
        this rate. It also corresponds to the maximum allowed deceleration.
        Usually doesn't need to be changed.
      t_shutoff: This class features an auto-shutoff thread that deenergizes
        the motor after a period of `t_shutoff` seconds of inactivity. The
        timer counts in steps of `0.1s`, which is thus the maximum precision
        for this setting. When set to `0`, this feature is disabled and the
        motor remains energized until the :meth:`close` method is called.
      config_file: The path to the config file to be loaded to the Tic. It only
        works if ``backend`` is 'ticcmd'. The config file contains some
        specific settings that can only be accessed this way using the 'ticcmd'
        backend. Not necessary for most applications.
      serial_number: The serial number of the Tic to be controlled. It must be
        given as a :obj:`str`, and it is an 8-digits number. Allows to control
        the right device if several Tic of the same model are connected.
        Otherwise, an error is raised.
      model: The model of the Tic to be controlled. Available models are:
        ::

          'T825', 'T824', 'T500', 'N825', 'T249', '36v4'

        Allows to control the right device if several Tic of different models
        are connected. Otherwise, an error is raised.
      reset_command_timeout: Enables or disables the `reset_command_timeout`
        thread. It can only be disabled if ``backend`` is 'USB'. This thread
        pings the Tic every `0.5s`, so that it doesn't block due to a Command
        Timeout error. This feature is a safety to prevent the motor from
        running indefinitely if the USB connection is down, so it is better not
        to disable it. When disabled, the Tic never raises Command Timeout
        errors.
      backend: The backend for communicating with the Tic. Available backends
        are:
        ::

          'USB', 'ticcmd'

        They both communicate over USB, but 'ticcmd' requires Pololu's firmware
        to be installed. Some features are specific to each backend.
      unrestricted_current_limit: Enables or disables the unrestricted current
        limit feature. Only works if ``backend`` is 'USB', and for the 36v4 Tic
        model. When disabled, the maximum current allowed is `3939mA`. If
        enabled, it goes up to `9095mA`. The Tic should however be cooled in
        order to withstand currents higher than `3939mA`.
      pin_function: Allows setting the Tic GPIO functions. It is a :obj:`dict`
        whose keys are the pin names, and values are the functions. Only works
        if ``backend`` is `'USB'`. Only the pins indicated in ``pin_function``
        are set, the others are left in their previous state. The available
        pins are:
        ::

          'SCL', 'SDA', 'TX', 'RX', 'RC'

        and can be set to:
        ::

          'Default', 'Kill switch', 'Limit switch forward', \
'Limit switch reverse'

        The GPIO functions remain set as long as they are not changed by the
        user, so for a given setup it is only necessary to set them once.
      pin_polarity: Allows setting the polarity of the GPIOs used as switches.
        It is a :obj:`dict`, whose keys are the pin names, and values are the
        pin polarities. Only works if ``backend`` is `'USB'`. Only the pins
        indicated in ``pin_function`` are set, the others are left in their
        previous state. The available pins are:
        ::

          'SCL', 'SDA', 'TX', 'RX', 'RC'

        and can be set to:
        ::

          'Active high', 'Active low'

        The GPIO polarities remain set as long as they are not changed by the
        user, so for a given setup it is only necessary to set them once.

    Warning:
      - ``current_limit``:
        If the ``current_limit`` setting is higher than the motor max current,
        there's a risk of overheating and damaging the motor !

    Note:
      - ``steps_per_mm``:
        If you have to measure this value, it can be done easily following this
        procedure. Set ``steps_per_mm`` to `spm` (`100` should be fine), and
        ``step_mode`` to `sm` (`8` should be fine). Run a Crappy script
        for moving the motor from position `0` to position `p` (a few tenth of
        millimeters should be fine). The motor will reach an actual position
        `ap` that can be measured. The actual ``steps_per_mm`` value `aspm` for
        this motor can be calculated as follows:
        ::

          aspm = spm * p / ap

      - ``step_mode``:
        Increasing the number of microsteps allows to reduce the noise, the
        vibrations, and improve the precision. However, the more microsteps,
        the lower the maximum achievable speed for the motor. Chances that the
        motor misses microsteps are also higher when the number of microsteps
        is high.
      - ``t_shutoff``:
        This functionality was originally added for long tests in temperature
        controlled environments, so that the motor doesn't unnecessarily heat
        the setup when inactive. In other assays, it may still be useful for
        reducing the noise, the electromagnetic interference, or the energy
        consumption.
      - ``serial_number``:
        Serial numbers can be accessed using the `lsusb` command in Linux
        shell, or running ``ticcmd --list`` if `ticcmd` is installed. This
        number is also displayed during :meth:`__init__` if only one device is
        connected and ``serial_number`` is :obj:`None`.
      - ``model``:
        The model is written on the Tic board, and can be accessed by running
        ``ticcmd --list`` in a shell if `ticcmd` is installed. It is also
        displayed during :meth:`__init__` if only one device is connected and
        ``model`` is :obj:`None`.
      - **Pins settings**:
        The pin functions and polarity can also be set independently of
        ``crappy`` before starting the test, in the `ticgui`.
    """

    super().__init__()

    if backend not in Tic_backends:
      raise ValueError("backend should be in {}".format(Tic_backends))
    else:
      self._backend = backend

    if model is not None and model not in Tic_product_id:
      raise ValueError("model should be in {} if given".format(list(
        Tic_product_id.keys())))

    if serial_number is not None and type(serial_number) is not str:
      raise ValueError("serial_number should be given as a string")

    # Finding the right device among all the connected ones
    if backend == 'USB':
      # Finding all devices matching the given inputs
      if model is None:
        if serial_number is None:
          devices = core.find(find_all=True,
                              idVendor=Tic_vendor_id)
        else:
          devices = core.find(find_all=True,
                              idVendor=Tic_vendor_id,
                              custom_match=FindSerialNumber(serial_number))
      else:
        if serial_number is None:
          devices = core.find(find_all=True,
                              idVendor=Tic_vendor_id,
                              idProduct=Tic_product_id[model])
        else:
          devices = core.find(find_all=True,
                              idVendor=Tic_vendor_id,
                              idProduct=Tic_product_id[model],
                              custom_match=FindSerialNumber(serial_number))
      # Making sure there's only one matching device
      devices = list(devices)
      if len(devices) == 0:
        raise IOError("No matching device connected")
      elif len(devices) > 1:
        raise IOError("Several matching devices found, try specifying a "
                      "device or a serial_number")
      else:
        self._dev = devices[0]
      # Setting self.serial_number and self.device
      if serial_number is None:
        self._serial_number = util.get_string(self._dev,
                                              self._dev.iSerialNumber)
      else:
        self._serial_number = serial_number
      if model is None:
        try:
          self._model = next(key for key, value in Tic_product_id.items() if
                             value == self._dev.idProduct)
        except StopIteration:
          raise ValueError("The Tic model automatically found is not "
                           "implemented in crappy")
      else:
        self._model = model

    elif backend == 'ticcmd':
      # Finding all devices matching the given inputs
      devices = check_output(['ticcmd', '--list']).decode("utf-8").split("\n")
      devices.pop()  # Removing the '' element at the end of devices
      devices = [string.split(',') for string in devices]
      if model is not None:
        if serial_number is not None:
          devices = [dev for dev in devices if dev[0] == serial_number and
                     model in dev[1]]
        else:
          devices = [dev for dev in devices if model in dev[1]]
      elif serial_number is not None:
        devices = [dev for dev in devices if dev[0] == serial_number]
      # Making sure there's only one matching device
      if len(devices) == 0:
        raise IOError("No matching device found")
      elif len(devices) > 1:
        raise IOError("Several matching devices found, try specifying a "
                      "device or a serial_number")
      # Setting self.serial_number and self.device
      if serial_number is None:
        self._serial_number = devices[0][0]
      else:
        self._serial_number = serial_number
      if model is None:
        try:
          self._model = next(key for key in Tic_product_id if key in
                             devices[0][1])
        except StopIteration:
          raise ValueError("The Tic model automatically found is not "
                           "implemented in crappy")
      else:
        self._model = model

    # Displaying model and serial_number if they were not specified by the user
    if serial_number is None:
      self.log(logging.INFO, f"Tic serial number: {self._serial_number}")
    if model is None:
      self.log(logging.INFO, f"Tic model: {self._model}")

    # Making sure the current limit is valid, especially for the 36v4 model
    if not 0 < current_limit < Tic_max_allowed_current[self._model]:
      if self._model == '36v4':
        if not 0 < current_limit < Tic_36v4_max_current:
          raise ValueError("current_limit should be between 0 and {} mA for "
                           "this Tic model".format(Tic_36v4_max_current))
        elif not unrestricted_current_limit:
          raise ValueError("current_limit exceeds the safety limit, which may "
                           "cause overheating. Set unrestricted_current_limit "
                           "to True if you want to keep this current_limit "
                           "(only works if backend='USB')")
        elif backend != 'USB':
          raise ValueError("Setting unrestricted_current_limit to True only "
                           "works if backend='USB'")
      else:
        raise ValueError("current limit should be between 0 and {} mA "
                         "for this Tic model".
                         format(Tic_max_allowed_current[self._model]))
    self._current_limit = current_limit
    self._unrestricted_current_limit = unrestricted_current_limit

    # Converting the current limit value to a current index, used by the
    # USB backend only
    if backend == 'USB':
      if self._model == 'T500':
        self._current_index = min(enumerate(Tic_current_tables[self._model]),
                                  key=lambda x: abs(x[1] - current_limit))[0]
      else:
        self._current_index = round(min(Tic_current_tables[self._model],
                                    key=lambda x: abs(x - current_limit)) /
                                    Tic_current_steps[self._model])

    if step_mode not in Tic_step_modes[self._model]:
      raise ValueError("step_mode should be in {}".format(
        Tic_step_modes[self._model]))
    else:
      self._step_mode = step_mode

    if steps_per_mm < 0:
      raise ValueError("steps_per_mm should be positive")
    else:
      self._steps_per_mm = steps_per_mm

    # Keeping the max_accel and max_decel values within the Tic ratings
    if max_accel < self._to_mm(Tic_min_accel / 100):
      self.log(logging.WARNING, "Requested acceleration below min allowed "
                                "acceleration, setting to min allowed "
                                "acceleration")
      max_accel = self._to_mm(Tic_min_accel / 100)
    elif max_accel > self._to_mm(Tic_max_accel / 100):
      self.log(logging.WARNING, "Requested acceleration exceeding max allowed "
                                "acceleration, setting to max allowed "
                                "acceleration")
      max_accel = self._to_mm(Tic_max_accel / 100)
    self._max_accel = max_accel

    if t_shutoff < 0:
      raise ValueError("t_shutoff should be zero or positive")
    else:
      self._t_shutoff = t_shutoff

    if config_file is not None and backend != 'ticcmd':
      self.log(logging.WARNING, "Config files can only be loaded if "
                                "backend='ticcmd', ignoring the given "
                                "config_file")
      self._config_file = None
    else:
      self._config_file = config_file

    if backend != 'USB' and not reset_command_timeout:
      self.log(logging.WARNING, "reset_command_timeout can only be disabled "
                                "if backend='USB', reset_command_timeout set "
                                "to True")
      self._rct_on = True
    else:
      self._rct_on = reset_command_timeout

    if backend != 'USB' and pin_function is not None:
      raise ValueError("It is not possible to set the pin functions if "
                       "the backend is not 'USB'")
    if pin_function is not None:
      if not all(key in Tic_pins_bit for key in pin_function):
        raise ValueError("Unexpected pin name, pin names should be in "
                         "{}".format(list(Tic_pins_bit.keys())))
      if not all(value in Tic_pin_modes for value in pin_function.values()):
        raise ValueError("Unexpected pin function, pin functions should be in "
                         "{}".format(list(Tic_pin_modes.keys())))
    self._pin_function = pin_function

    if backend != 'USB' and pin_polarity is not None:
      raise ValueError("It is not possible to set the pin polarities if "
                       "the backend is not 'USB'")
    if pin_polarity is not None:
      if not all(key in Tic_pins_bit for key in pin_polarity):
        raise ValueError("Unexpected pin name, pin names should be in "
                         "{}".format(list(Tic_pins_bit.keys())))
      if not all(value in Tic_pin_polarity for value in pin_polarity.values()):
        raise ValueError("Unexpected pin function, pin functions should be in "
                         "{}".format(list(Tic_pin_polarity.keys())))
    self._pin_polarity = pin_polarity

    # The lock is meant for preventing interferences between the threads
    self._lock = RLock()

    # Definition of the flags
    self._timer_shutoff = False
    self._RCT = False
    self._reset_timer_shutoff = False
    self._close = False

    # Definition of the auxiliary threads
    self._thrd_rct = Thread(target=self._thread_rct)
    self._thrd_shutoff = Thread(target=self._thread_shutoff)

  def open(self) -> None:
    """Sets the communication, the motor parameters and starts the threads."""

    if self._backend == 'USB':
      try:
        self.log(logging.INFO, f"Setting configuration on USB device "
                               f"{self._dev}")
        self._dev.set_configuration()
      except core.USBError:
        self.log(logging.ERROR,
                 "An error occurred while setting the configuration of the USB"
                 " device !\nYou may have to install the udev-rules for this "
                 "USB device, this can be done using the udev_rule_setter "
                 "utility in the util folder")
        raise

    # Setting the Tic according to the user parameters
    pin_changed = False
    if self._pin_polarity is not None:
      self._set_pin_polarity(self._pin_polarity)
      pin_changed = True
    if self._pin_function is not None:
      self._set_pin_function(self._pin_function)
      pin_changed = True
    if pin_changed:
      self._reset()

    self._enter_safe_start()
    self._deenergize()
    self._set_step_mode()
    self._set_current_limit()
    self._set_max_accel()
    self._set_max_decel()

    # Loading the config file
    if self._config_file is not None:
      self.log(logging.INFO, "Loading the config file")
      self._ticcmd('--settings', str(self._config_file))

    # Starting the auxiliary threads
    # The RCT thread is not needed in case reset_command_timeout is False
    # The shutoff thread is not needed in case t_shutoff is zero
    if self._rct_on:
      self.log(logging.INFO, "Starting the thread for managing the command "
                             "timeouts")
      self._thrd_rct.start()
    else:
      self.log(logging.INFO, "Disabling the command timeout security")
      self._usb_command(request=Tic_cmd['Set_setting'],
                        value=0x00,
                        index=Tic_settings['Command_timeout_low'])
      self._usb_command(request=Tic_cmd['Set_setting'],
                        value=0x00,
                        index=Tic_settings['Command_timeout_high'])
    if self._t_shutoff > 0:
      self.log(logging.INFO, "Starting the thread managing the auto shutoff")
      self._thrd_shutoff.start()

  def get_speed(self) -> float:
    """Reads the current motor speed.

        Returns:
          The speed in `mm/s`
        """

    if self._backend == 'ticcmd':
      return self._to_mm(yaml.load(self._ticcmd('-s'), Loader=yaml.FullLoader)
                         ['Current velocity'] / 10000) if full_loader else \
        self._to_mm(yaml.load(self._ticcmd('-s'))['Current velocity'] / 10000)
    elif self._backend == 'USB':
      return self._to_mm(int.from_bytes(
        self._usb_command(request_type=Tic_usb_request['Var'],
                          request=Tic_cmd['Get_variable'],
                          index=Tic_var['Current_velocity'],
                          data_or_length=4),
        byteorder='little',
        signed=True) / 10000)

  def get_position(self) -> float:
    """Reads the current motor position.

    Returns:
      The position in `mm`
    
    .. versionchanged:: 1.5.2 renamed from get_pos to get_position
    """

    if self._backend == 'ticcmd':
      return self._to_mm(yaml.load(self._ticcmd('-s'), Loader=yaml.FullLoader)
                         ['Current position']) if full_loader else \
        self._to_mm(yaml.load(self._ticcmd('-s'))['Current position'])
    elif self._backend == 'USB':
      return self._to_mm(int.from_bytes(
        self._usb_command(request_type=Tic_usb_request['Var'],
                          request=Tic_cmd['Get_variable'],
                          index=Tic_var['Current_position'],
                          data_or_length=4),
        byteorder='little',
        signed=True))

  def set_position(self, position: float, speed: Optional[float]) -> None:
    """Sends a position command to the motor.

    Args:
      position: The position to reach in `mm`
      speed: The speed at which the motor should move to the given position,
        in `mm/s` Giving a speed other than :obj:`None` will set the maximum
        speed of the motor to that speed.

    Note:
      - ``speed``:
        The only way to reach a position at a given speed is to change the
        maximum speed. The Tic will try to accelerate to the maximum speed but
        may remain slower if it doesn't have time to do so before reaching the
        given position.
    
    .. versionchanged:: 2.0.0 *speed* is now a mandatory argument
    """

    if speed is not None:
      self._set_max_speed(speed)

    # Energizing the motor
    self._energize()
    self._exit_safe_start()

    # Raising the flags
    self._timer_shutoff = True
    self._reset_timer_shutoff = True
    self._RCT = True

    # Sending the position command
    self._set_position(position)

  def set_speed(self, speed: float) -> None:
    """Sends a speed command to the motor.

    Args:
      speed: The speed the motor should reach
    """

    # Changing the maximum speed if needed
    max_speed = self._get_max_speed()
    if abs(speed) > max_speed:
      self._set_max_speed(speed)

    # The command speed may first need to be reduced or increased in order to
    # comply with the Tic ratings
    final_speed = min(abs(self._to_steps(speed * 10000)), Tic_max_speed)
    if final_speed:  # If speed is 0, then it should remain 0
      if final_speed < Tic_min_speed:
        self.log(logging.WARNING, "Requested speed below min possible speed, "
                                  "setting to min possible speed")
        final_speed = Tic_min_speed
      final_speed *= speed / abs(speed)

    # Energizing the motor
    self._energize()
    self._exit_safe_start()

    # Raising the flags
    self._timer_shutoff = True
    self._reset_timer_shutoff = True
    self._RCT = True

    # Sending the speed command
    self._set_velocity(final_speed)

  def stop(self) -> None:
    """Sets the speed to `0`."""

    self._set_velocity(0)

  def close(self) -> None:
    """Stops the motor, joins the threads and deenergizes the motor."""

    self._close = True
    if self._rct_on:
      self.log(logging.INFO, "Stopping the thread managing the command "
                             "timeouts")
      self._thrd_rct.join()
    if self._t_shutoff > 0:
      self.log(logging.INFO, "Starting the thread managing the auto shutoff")
      self._thrd_shutoff.join()
    self._enter_safe_start()
    self._deenergize()
    if self._backend == 'USB':
      self.log(logging.INFO, "Releasing the USB resources")
      util.dispose_resources(self._dev)

  def _to_steps(self, mm: float) -> float:
    """Wrapper for converting `mm` to `steps`."""

    return mm * self._steps_per_mm * self._step_mode

  def _to_mm(self, steps: float) -> float:
    """Wrapper for converting `steps` to `mm`."""

    return steps / self._steps_per_mm / self._step_mode

  def _reset_command_timeout(self) -> None:
    """Sends a reset command timeout command."""

    self.log(logging.DEBUG, "Sending reset command timeout command")

    if self._backend == 'ticcmd':
      self._ticcmd('--reset-command-timeout')
    elif self._backend == 'USB':
      self._usb_command(request=Tic_cmd['Reset_command_timeout'])

  def _enter_safe_start(self) -> None:
    """Sends an enter safe start command."""

    self.log(logging.INFO, "Entering safe start mode")

    if self._backend == 'ticcmd':
      self._ticcmd('--enter-safe-start')
    elif self._backend == 'USB':
      self._usb_command(request=Tic_cmd['Enter_safe_start'])

  def _exit_safe_start(self) -> None:
    """Sends an exit safe start command."""

    self.log(logging.INFO, "Exiting safe start mode")

    if self._backend == 'ticcmd':
      self._ticcmd('--exit-safe-start')
    elif self._backend == 'USB':
      self._usb_command(request=Tic_cmd['Exit_safe_start'])

  def _deenergize(self) -> None:
    """Sends a deenergize command."""

    self.log(logging.INFO, "Deenergizing the motor")

    if self._backend == 'ticcmd':
      self._ticcmd('--deenergize')
    elif self._backend == 'USB':
      self._usb_command(request=Tic_cmd['Deenergize'])

  def _energize(self) -> None:
    """Sends an energize command."""

    self.log(logging.INFO, "Energizing the motor")

    if self._backend == 'ticcmd':
      self._ticcmd('--energize')
    elif self._backend == 'USB':
      self._usb_command(request=Tic_cmd['Energize'])

  def _set_step_mode(self) -> None:
    """Sends a set step mode command."""

    self.log(logging.INFO, "Setting the micro-step mode")

    if self._backend == 'ticcmd':
      self._ticcmd('--step-mode', str(self._step_mode))
    elif self._backend == 'USB':
      self._usb_command(request=Tic_cmd['Set_step_mode'],
                        value=Tic_step_mode[self._step_mode])

  def _set_current_limit(self) -> None:
    """Sends a set current limit command."""

    self.log(logging.INFO, "Setting the current limit")

    if self._backend == 'ticcmd':
      self._ticcmd('--current', str(self._current_limit))
    elif self._backend == 'USB':
      if self._model == '36v4':
        self._usb_command(request=Tic_cmd['Set_setting'],
                          value=self._unrestricted_current_limit,
                          index=Tic_settings['Unrestricted_current_limit'])
      self._usb_command(request=Tic_cmd['Set_current_limit'],
                        value=self._current_index)

  def _set_max_speed(self, speed: float) -> None:
    """Clamps the speed within the limits and sets it."""

    self.log(logging.DEBUG, "Setting the maximum speed")

    # The given speed may first need to be reduced or increased in order to
    # comply with the Tic ratings
    if abs(speed) > self._to_mm(Tic_max_speed / 10000):
      self.log(logging.WARNING, "Requested speed exceeding max allowed speed, "
                                "setting to max allowed speed")
      max_speed = Tic_max_speed
    elif abs(speed) < self._to_mm(Tic_min_speed / 10000):
      self.log(logging.WARNING, "Requested speed below min possible speed, "
                                "setting to min possible speed")
      max_speed = Tic_min_speed
    else:
      max_speed = abs(self._to_steps(speed * 10000))

    if self._backend == 'ticcmd':
      self._ticcmd('--max-speed', str(int(max_speed)))
    elif self._backend == 'USB':
      self._usb_32_bit(request=Tic_cmd['Set_max_speed'],
                       data=int(max_speed))

  def _set_max_accel(self) -> None:
    """Clamps the acceleration within the limits and sets it."""

    self.log(logging.INFO, "Setting the maximum acceleration")

    if self._backend == 'ticcmd':
      self._ticcmd('--max-accel', str(int(
        self._to_steps(self._max_accel * 100))))
    elif self._backend == 'USB':
      self._usb_32_bit(request=Tic_cmd['Set_max_accel'],
                       data=int(self._to_steps(self._max_accel * 100)))

  def _set_max_decel(self) -> None:
    """Clamps the deceleration within the limits and sets it."""

    self.log(logging.INFO, "Setting the maximum deceleration")

    if self._backend == 'ticcmd':
      self._ticcmd('--max-decel', str(int(
        self._to_steps(self._max_accel * 100))))
    elif self._backend == 'USB':
      self._usb_32_bit(request=Tic_cmd['Set_max_decel'],
                       data=int(self._to_steps(self._max_accel * 100)))

  def _set_position(self, position: float) -> None:
    """Sends a set position command."""

    self.log(logging.DEBUG, "Setting the position")

    if self._backend == 'ticcmd':
      self._ticcmd('--position', str(int(self._to_steps(position))))
    elif self._backend == 'USB':
      self._usb_32_bit(request=Tic_cmd['Set_target_position'],
                       data=int(self._to_steps(position)))

  def _set_velocity(self, velocity: float) -> None:
    """Sends a set velocity command."""

    self.log(logging.DEBUG, "Setting the velocity")

    if self._backend == 'ticcmd':
      self._ticcmd('--velocity', str(int(velocity)))
    elif self._backend == 'USB':
      self._usb_32_bit(request=Tic_cmd['Set_target_velocity'],
                       data=int(velocity))

  def _get_max_speed(self) -> float:
    """Reads the maximum speed from the motor."""

    self.log(logging.DEBUG, "Reading the maximum speed")

    if self._backend == 'ticcmd':
      return self._to_mm(yaml.load(self._ticcmd('-s', '--full'),
                                   Loader=yaml.FullLoader)
                         ['Max speed'] / 10000) if full_loader else \
        self._to_mm(yaml.load(self._ticcmd('-s', '--full'))
                    ['Max speed'] / 10000)
    elif self._backend == 'USB':
      return self._to_mm(int.from_bytes(
          self._usb_command(request_type=Tic_usb_request['Var'],
                            request=Tic_cmd['Get_variable'],
                            index=Tic_var['Max_speed'],
                            data_or_length=4),
          byteorder='little',
          signed=False) / 10000)

  def _set_pin_function(self, pin_func: Dict[str, str]) -> None:
    """Sets the pin function bitfields.

    Sends a command for setting each pin separately, and three commands for
    setting the Kill switch, Limit switch forward and Limit switch reverse
    bitfields.
    """

    self.log(logging.INFO, "Setting the pin functions")

    if self._backend == 'ticcmd':
      pass
    elif self._backend == 'USB':
      # Reads the current bitfields
      kill_switch_map = int.from_bytes(
        self._usb_command(request_type=Tic_usb_request['Var'],
                          request=Tic_cmd['Get_setting'],
                          index=Tic_settings['Kill_switch_map'],
                          data_or_length=1),
        byteorder='little',
        signed=False)
      limit_forward_map = int.from_bytes(
        self._usb_command(request_type=Tic_usb_request['Var'],
                          request=Tic_cmd['Get_setting'],
                          index=Tic_settings['Limit_switch_forward_map'],
                          data_or_length=1),
        byteorder='little',
        signed=False)
      limit_reverse_map = int.from_bytes(
        self._usb_command(request_type=Tic_usb_request['Var'],
                          request=Tic_cmd['Get_setting'],
                          index=Tic_settings['Limit_switch_reverse_map'],
                          data_or_length=1),
        byteorder='little',
        signed=False)
      for pin in pin_func:
        # Modifies the three bitfields
        if pin_func[pin] == 'Default':
          kill_switch_map &= 0xFF - (1 << Tic_pins_bit[pin])
          limit_forward_map &= 0xFF - (1 << Tic_pins_bit[pin])
          limit_reverse_map &= 0xFF - (1 << Tic_pins_bit[pin])
        elif pin_func[pin] == 'Kill switch':
          kill_switch_map |= 1 << Tic_pins_bit[pin]
          limit_forward_map &= 0xFF - (1 << Tic_pins_bit[pin])
          limit_reverse_map &= 0xFF - (1 << Tic_pins_bit[pin])
        elif pin_func[pin] == 'Limit switch forward':
          kill_switch_map &= 0xFF - (1 << Tic_pins_bit[pin])
          limit_forward_map |= 1 << Tic_pins_bit[pin]
          limit_reverse_map &= 0xFF - (1 << Tic_pins_bit[pin])
        elif pin_func[pin] == 'Limit switch reverse':
          kill_switch_map &= 0xFF - (1 << Tic_pins_bit[pin])
          limit_forward_map &= 0xFF - (1 << Tic_pins_bit[pin])
          limit_reverse_map |= 1 << Tic_pins_bit[pin]

        # Sends an individual command for each pin
        self._usb_command(request=Tic_cmd['Set_setting'],
                          value=Tic_pin_modes[pin_func[pin]],
                          index=Tic_settings[pin + '_config'])
      # Sets the three bitfields
      self._usb_command(request=Tic_cmd['Set_setting'],
                        value=kill_switch_map,
                        index=Tic_settings['Kill_switch_map'])
      self._usb_command(request=Tic_cmd['Set_setting'],
                        value=limit_forward_map,
                        index=Tic_settings['Limit_switch_forward_map'])
      self._usb_command(request=Tic_cmd['Set_setting'],
                        value=limit_reverse_map,
                        index=Tic_settings['Limit_switch_reverse_map'])

  def _set_pin_polarity(self, pin_pol: Dict[str, str]) -> None:
    """Sets the switch polarity bitfield."""

    self.log(logging.INFO, "Setting the pin polarities")

    if self._backend == 'ticcmd':
      pass
    elif self._backend == 'USB':
      current = int.from_bytes(
        self._usb_command(request_type=Tic_usb_request['Var'],
                          request=Tic_cmd['Get_setting'],
                          index=Tic_settings['Switch_polarity_map'],
                          data_or_length=1),
        byteorder='little',
        signed=False)
      for pin in pin_pol:
        if Tic_pin_polarity[pin_pol[pin]]:
          current |= 1 << Tic_pins_bit[pin]
        else:
          current &= 0xFF - (1 << Tic_pins_bit[pin])
      self._usb_command(request=Tic_cmd['Set_setting'],
                        value=current,
                        index=Tic_settings['Switch_polarity_map'])

  def _reset(self) -> None:
    """Resets the Tic and reloads the settings."""

    self.log(logging.INFO, "Resetting the device")

    if self._backend == 'ticcmd':
      pass
    elif self._backend == 'USB':
      self._usb_command(request=Tic_cmd['Reset'])

  def _usb_command(self,
                   request_type: int = Tic_usb_request['Cmd'],
                   request: int = 0,
                   value: int = 0,
                   index: int = 0,
                   data_or_length: int = 0) -> Union[bytearray, int]:
    """Wrapper for sending a USB control transfer."""

    with self._lock:
      try:
        self.log(logging.DEBUG, f"Sending USB command with "
                                f"request type {request_type}, request "
                                f"{request}, value {value}, index {index},"
                                f"length or data {data_or_length}")
        result = self._dev.ctrl_transfer(bmRequestType=request_type,
                                         bRequest=request,
                                         wValue=value,
                                         wIndex=index,
                                         data_or_wLength=data_or_length)
      except core.USBError:
        raise IOError("An error occurred during USB communication")
    return result

  def _usb_32_bit(self, request: int, data: int) -> None:
    """Wrapper for sending USB requests containing 32-bits values."""

    value = data & 0xFFFF
    index = data >> 16 & 0xFFFF
    self._usb_command(request=request, value=value, index=index)

  def _ticcmd(self, *args: str) -> bytes:
    """Wrapper for calling ticcmd in a subprocess."""

    with self._lock:
      self.log(logging.DEBUG, f"Calling ticcmd -d {self._serial_number} "
                              f"{' '.join(args)}")
      return check_output(['ticcmd'] + ['-d'] +
                          [self._serial_number]
                          + list(args))

  def _thread_shutoff(self) -> None:
    """Thread for deenergizing the motor after a given period of inactivity.

    This thread reads the speed every 0.1s, and increments a timer if the speed
    is `0`. Once the timer reaches `_t_shutoff`, deenergizes the motor. The
    timer is reset if a speed or position command is issued, or if the speed is
    not `0`.
    """

    timer = 0
    while not self._close:
      sleep(0.01)

      while self._timer_shutoff:
        # Exit if close flag raised
        if self._close:
          break

        # Resetting timer if reset flag raised
        if self._reset_timer_shutoff:
          timer = 0
          self._reset_timer_shutoff = False

        # Checking if the motor is moving
        if self.get_speed() == 0:
          timer += 0.1
        else:
          timer = 0
        sleep(0.1)

        # Finally deenergizing the motor if all the conditions are met
        if timer > self._t_shutoff and \
                not self._reset_timer_shutoff and \
                not self._close:
          self._enter_safe_start()
          self._deenergize()
          self._timer_shutoff = False
          self._RCT = False  # Stopping the RCT thread as well
          timer = 0

  def _thread_rct(self) -> None:
    """Thread for sending the reset command timeout command every `0.5s`.

    This prevents the motor from stopping because of a reset command timeout
    error. Only sends the command when the motor is energized.
    """

    # Setting command timeout to 1000ms
    if self._backend == 'USB':
      self._usb_command(request=Tic_cmd['Set_setting'],
                        value=0xE8,
                        index=Tic_settings['Command_timeout_low'])
      self._usb_command(request=Tic_cmd['Set_setting'],
                        value=0x03,
                        index=Tic_settings['Command_timeout_high'])

    while not self._close:
      sleep(0.01)
      while self._RCT:
        if self._close:
          break
        self._reset_command_timeout()
        sleep(0.5)



# coding: utf-8

from .camera_config import CameraConfig
from .camera_config_boxes import CameraConfigBoxes
from .config_tools import Box, Overlay, SpotsBoxes, SpotsDetector, Zoom
from .dic_ve_config import DICVEConfig
from .dis_correl_config import DISCorrelConfig
from .video_extenso_config import VideoExtensoConfig



# coding: utf-8

from time import sleep
from re import findall
from typing import Optional
import logging
from  warnings import warn

from .meta_actuator import Actuator
from .._global import OptionalModule

try:
  from serial import Serial
except (ModuleNotFoundError, ImportError):
  Serial = OptionalModule('pyserial')


class NewportTRA6PPD(Actuator):
  """Drives the Newport TRA6PPD linear actuator in position.

  Warning:
    This actuator cannot handle a high serial messages rate. It is recommended
    to set the frequency of the corresponding :class:`~crappy.blocks.Machine`
    Block to a few dozen `Hz` at most.

  Note:
    This Actuator ignores new position commands while it is moving.
  
  .. versionadded:: 1.5.10
  .. versionchanged:: 2.0.0 renamed from TRA6PPD to NewportTRA6PPD
  """

  def __init__(self,
               baudrate: int = 57600,
               port: str = '/dev/ttyUSB0') -> None:
    """Sets the instance attributes and initializes the parent class.

    Args:
      baudrate: The baudrate for the serial connection.
      port: Path to the port to use for serial communication.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._ser = None

    super().__init__()

    self._max_speed = 0.2
    self._min_position = 0
    self._max_position = 6

    self._baudrate = baudrate
    self._port = port

  def open(self) -> None:
    """Resets the device and performs homing."""

    self.log(logging.INFO, f"Opening the serial port {self._port} with "
                           f"baudrate {self._baudrate}")
    self._ser = Serial(self._port, baudrate=self._baudrate, timeout=0.1)

    # First, reset the device
    self.log(logging.DEBUG, f"Writing b'1RS\\r\\n' to port {self._port}")
    self._ser.write(b'1RS\r\n')
    sleep(5)

    # Then, perform homing (may take up to 45s)
    self.log(logging.DEBUG, f"Writing b'10R\\r\\n' to port {self._port}")
    self._ser.write(b'1OR\r\n')
    for i in range(5, 0, -1):
      self.log(logging.INFO, f"Performing homing, {10 * i} seconds left")
      sleep(10)

  def set_position(self, position: float, speed: Optional[float]) -> None:
    """Sends the actuator a command to reach a given position.

    The command is ignored if the actuator is already moving.

    Args:
      position: The position to reach. Should be between `0` and `6 mm`.
      speed: The speed at which the actuator should move. Should be between `0`
        and `0.2 mm/s`. If :obj:`None` is received, the default is `0.2 mm/s`.
    """

    if speed is None:
      speed = 0.2

    # Clamping the speed command into the allowed interval
    speed_clamped = max(min(self._max_speed, speed), 0)

    # Sending the speed command
    self.log(logging.DEBUG, f"Writing b'1VA{speed_clamped:.5f}\\r\\n' to port "
                            f"{self._port}")
    self._ser.write(f'1VA{speed_clamped:.5f}\r\n'.encode())

    # Ignoring the position command if it is out of range
    if not self._min_position <= position <= self._max_position:
      self.log(logging.WARNING, f"The requested position {position} is out of "
                                f"range ! Ignoring")
    else:
      # Sending the position command
      self.log(logging.DEBUG, f"Writing b'1PA{position:.5f}\\r\\n' to port "
                              f"{self._port}")
      self._ser.write(f'1PA{position:.5f}\r\n'.encode())

  def get_position(self) -> float:
    """Reads the current position.

    Returns:
      Current position of the motor.
    """

    # Sending the read command
    self.log(logging.DEBUG, f"Writing b'1TP?\\r\\n' to port {self._port}")
    self._ser.write(b'1TP?\r\n')

    # Using a regular expression to parse the answer
    ret = self._ser.readline()
    self.log(logging.DEBUG, f"Read {ret} on port {self._port}")
    return float(findall(r'\d\.\d+', ret.decode())[0])

  def close(self) -> None:
    """Closes the serial port."""

    if self._ser is not None:
      self.log(logging.INFO, f"Closing the serial port {self._port}")
      self._ser.close()

  def stop(self) -> None:
    """Stops the motor and sets the device to "disable" state."""

    if self._ser is not None:
      self.log(logging.DEBUG, f"Writing b'ST\\r\\n' to port {self._port}")
      self._ser.write(b'ST\r\n')
      self.log(logging.DEBUG, f"Writing b'1MMO\\r\\n' to port {self._port}")
      self._ser.write(b'1MMO\r\n')
